# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union
import numpy as np

from rarity.data_loader import CSVDataLoader, DataframeLoader


class BaseInterpreters:
    '''
    Class to tapout base dataframe that combines all xfeatures, yTrue and yPred that fits for next level of input transformation
    based on the integration need of specific feature component.

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module
    '''
    def __init__(self, data_loader: Union[CSVDataLoader, DataframeLoader]):
        self.data_loader = data_loader
        self.analysis_type = data_loader.get_analysis_type()
        self.models = data_loader.get_model_list()

    def get_df_with_offset_values(self):
        '''
        For use in regression task only.
        Include offset values into main combined df collating together with xFeatures, yTrue and yPred

        Returns:
            :obj:`~pandas.DataFrame`
        '''
        df = self.data_loader.get_all()
        df[f'offset_{self.models[0]}'] = df[f'yPred_{self.models[0]}'] - df['yTrue']
        if len(self.models) > 1:
            df[f'offset_{self.models[1]}'] = df[f'yPred_{self.models[1]}'] - df['yTrue']            
        return df

    def get_df_with_probability_values(self):
        '''
        For use in classification task only.
        Include prediction state indicating if prediction is correct or miss-predict into main combined df
        collating together with yTrue and probabilities value of each class label

        Returns :
            ls_dfs_viz (:obj:`List[~pd.DataFrame]`):
                List of dfs with prediction state info included for each model
            ls_class_labels (:obj:`List[str]`):
                List of class labels
        '''
        # extract list of class labels
        df_tmp = self.data_loader.get_all()[0]
        label_start_idx = list(df_tmp.columns).index('yTrue') + 1
        label_stop_idx = list(df_tmp.columns).index('model')
        ls_class_labels = list(df_tmp.columns)[label_start_idx:label_stop_idx]

        ls_dfs_viz = []
        for df in self.data_loader.get_all():
            # extract df with info useful for miss-prediction viz
            col_start_idx_to_extract = list(df.columns).index('yTrue')
            df_viz = df.loc[:, list(df.columns)[col_start_idx_to_extract:]]
            pred_state_condition = df_viz['yPred-label'].astype('str') == df_viz['yTrue'].astype('str')
            df_viz['pred_state'] = np.where(pred_state_condition, 'correct', 'miss-predict')

            # re-arrange the columns orders for viz need at features-module
            org_cols = list(df_viz.columns)
            org_cols.remove('model')
            new_position_yTrue_col_idx = org_cols.index('yPred-label')
            new_cols_order = ['model'] + org_cols[1:new_position_yTrue_col_idx] + ['yTrue'] + org_cols[new_position_yTrue_col_idx:]
            df_viz_final = df_viz.loc[:, new_cols_order]
            ls_dfs_viz.append(df_viz_final)
        return ls_dfs_viz, ls_class_labels



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict
import plotly.express as px


def plot_distribution_by_specific_feature(ls_specific_feature: List[str], kl_div_dict_sorted: Dict, comparison_base: str, model_name: str):
    '''
    Create distribution plot for a specific feature

    Arguments:
        ls_specific_feature (:obj:`List[str]`):
            list of feature to have its distribution graph plotted
        kl_div_dict_sorted (:obj:`Dict`):
            dictionary storing kl-divergence score by feature in decending order
        comparison_base (str):
            info to indicate the baseline for distribution comparison. ``dataset_type`` for regression and ``pred_state`` for classification task
        model_name (str):
            model used to generate yPred

    Returns:
        :obj:`List[~plotly.graph_objects.Figure]`:
            List of figures displaying distribution plot of specific feature
    '''
    if not isinstance(ls_specific_feature, list):
        ls_specific_feature = [ls_specific_feature]

    fig_obj_ls = []
    for specific_feature in ls_specific_feature:
        fig = _single_dist_plot(specific_feature, kl_div_dict_sorted, comparison_base, model_name)
        fig_obj_ls.append(fig)
    return fig_obj_ls


def plot_distribution_by_kl_div_ranking(kl_div_dict_sorted: Dict, display_option: str, display_value: int,
                                        comparison_base: str, model_name: str):
    '''
    Create distribution plot by kl-divergence score ranking in descending order

    Arguments:
        kl_div_dict_sorted (:obj:`Dict`):
            dictionary storing kl-divergence score by feature in decending order
        display_option (str)
            - info to indicate if to display distribution plot by top-N / bottom-N or both top-N + bottom-N
            - Available options: ``top``, ``bottom`` or ``both``
        display_value (int)
            - number indicates the limit of graph to be displayed, max at 10
            - if dataset consists of < 10 features, the limit == no. of features the dataset has
        comparison_base (str):
            info to indicate the baseline for distribution comparison. ``dataset_type`` for regression and ``pred_state`` for classification task
        model_name (str):
            model used to generate yPred

    Returns:
        :obj:`Dict[str, ~plotly.graph_objects.Figure]`:
            Dictionary storing distribution figures by display_option
    '''
    fig_obj_dict = {}

    if display_option == 'top':
        sliced_kl_div_dict_top = dict(list(kl_div_dict_sorted.items())[:display_value])
        fig_obj_dict['top'] = [plot_distribution_by_specific_feature(k, sliced_kl_div_dict_top, comparison_base, model_name)
                                for k, v in sliced_kl_div_dict_top.items()]
        fig_obj_dict['bottom'] = []

    elif display_option == 'bottom':
        sliced_kl_div_dict_btm = dict(list(kl_div_dict_sorted.items())[-display_value:])
        fig_obj_dict['top'] = []
        fig_obj_dict['bottom'] = [plot_distribution_by_specific_feature(k, sliced_kl_div_dict_btm, comparison_base, model_name)
                                    for k, v in sliced_kl_div_dict_btm.items()]

    elif display_option == 'both':
        sliced_kl_div_dict_top = dict(list(kl_div_dict_sorted.items())[:display_value])
        sliced_kl_div_dict_btm = dict(list(kl_div_dict_sorted.items())[-display_value:])
        fig_obj_dict['top'] = [plot_distribution_by_specific_feature(k, sliced_kl_div_dict_top, comparison_base, model_name)
                                for k, v in sliced_kl_div_dict_top.items()]
        fig_obj_dict['bottom'] = [plot_distribution_by_specific_feature(k, sliced_kl_div_dict_btm, comparison_base, model_name)
                                    for k, v in sliced_kl_div_dict_btm.items()]
    return fig_obj_dict


def _single_dist_plot(feature: str, kl_div_dict_sorted: Dict, comparison_base: str, model_name: str):
    '''
    Internal function to plot single distribution graph

    Important Arguments:

        comparison_base (str):
        ``dataset_type`` for regression, ``pred_state`` for classification
    '''
    df = kl_div_dict_sorted[feature][1]

    fig = px.histogram(df, x=feature, color=comparison_base,
                        marginal='box', opacity=0.5, barmode='overlay',
                        # color_discrete_sequence=px.colors.qualitative.D3)
                        # replaces default color mapping by value
                        color_discrete_map={'correct': '#1F77B4', 'miss-predict': '#FF7F0E', 'df_reference': '#1F77B4', 'df_sliced': '#FF7F0E'},
                        category_orders={'pred_state': ['correct', 'miss-predict'], 'dataset_type': ['df_reference', 'df_sliced']})

    kl_div_score = kl_div_dict_sorted[feature][0]

    customized_title = f'<b>[ KL divergence : {kl_div_score:.4f} ]</b><br>Distribution of xFeature : {feature}'
    customized_margin = dict(t=120)
    if model_name is not None:
        model_name = f'<span style="color:blue; font-size:14px">{model_name}   </span>'
        feature_name = f'<span style="font-size:14px">Distribution of xFeature : {feature}</span>'
        customized_title = f'<b>[ KL divergence : {kl_div_score:.4f} ]</b><br>' + model_name + feature_name
        customized_margin = dict(t=130)

    fig.update_layout(
            title=customized_title,
            xaxis_title=f'xFeature : {feature}',
            yaxis_title='count',
            title_x=0.48,
            legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1, title=None),
            margin=customized_margin)
    return fig



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

import dash
from dash.dependencies import Input, Output, ALL
from dash.exceptions import PreventUpdate
import dash_core_components as dcc
import dash_html_components as html
import dash_bootstrap_components as dbc

from rarity.app import app
from rarity.data_loader import CSVDataLoader, DataframeLoader
from rarity.interpreters.structured_data import IntMissPredictions
from rarity.visualizers import miss_predictions as viz_misspred
from rarity.visualizers import shared_viz_component as viz_shared
from rarity.utils import style_configs
from rarity.utils.common_functions import (identify_active_trace, is_active_trace, is_reset, is_regression, is_classification,
                                            detected_legend_filtration, detected_unique_figure, detected_more_than_1_unique_figure,
                                            detected_single_xaxis, detected_single_yaxis, get_min_max_index, get_min_max_offset,
                                            get_adjusted_xy_coordinate, conditional_sliced_df, dataframe_prep_on_model_count_by_yaxis_slice,
                                            insert_index_col)


def fig_plot_prediction_offset_overview(data_loader: Union[CSVDataLoader, DataframeLoader]):
    '''
    For use in regression task only.
    Display scatter plot for overview on prediction offset values

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying scatter plot outlining overview on prediction offset values by index
    '''
    df = IntMissPredictions(data_loader).xform()
    fig_obj = viz_misspred.plot_prediction_offset_overview(df)
    return fig_obj, df


def fig_probabilities_spread_pattern(data_loader: Union[CSVDataLoader, DataframeLoader]):
    '''
    For use in classification task only.
    Function to output collated info packs used to display final graph objects and data tables

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Returns:

            Compact outputs consist of the followings

            - fig_objs_all_models (:obj: `List[~plotly.graph_objects.Figure]`): figure displaying scatter plot outlining probabilities \
                comparison on correct data point vs miss-predicted data point for each class label
            - tables_all_models (:obj:`List[~dash_table.DataTable]`): table object outlining simple stats on ss, %correct, % wrong, accuracy \
                for each label class
            - ls_dfs_viz (:obj:`List[~pandas.DataFrame]`): dataframes for overview visualization need with true labels and \
                predicted labels included
            - df_features (:obj:`~pandas.DataFrame`): dataframe storing all features used in dataset
            - ls_class_labels (:obj:`List[str]`): list of class labels found in dataset
    '''
    ls_dfs_viz, ls_class_labels, ls_dfs_by_label, ls_dfs_by_label_state = IntMissPredictions(data_loader).xform()

    # prepare fig-objs and corresponding sub-tables
    fig_objs_all_models = []
    tables_all_models = []
    for i, ls_dfs in enumerate(ls_dfs_by_label):
        fig_objs_per_model = []
        tables_per_model = []
        for j, df_specific_label in enumerate(ls_dfs):
            fig_obj = viz_misspred.plot_probabilities_spread_pattern(df_specific_label)
            fig_objs_per_model.append(fig_obj)

            table_j = viz_misspred.plot_simple_probs_spread_overview(ls_dfs_by_label_state[i][j])
            tables_per_model.append(table_j)
        fig_objs_all_models.append(fig_objs_per_model)
        tables_all_models.append(tables_per_model)

    # prepare feature table
    df_features = data_loader.get_features()
    return fig_objs_all_models, tables_all_models, ls_dfs_viz, df_features, ls_class_labels


def table_with_relayout_datapoints(data, customized_cols, header, exp_format):
    '''
    Create table outlining dataframe content

    Arguments:
        data (:obj:`~dash_table.DataTable`):
            dictionary like format storing dataframe info under 'record' key
        customized_cols (:obj:`List[str]`):
            list of customized column names
        header (:obj:`Dict`):
            dictionary format storing the style info for table header
        exp_format (str):
            text info indicating the export format

    Returns:
        :obj:`~dash_table.DataTable`:
            table object outlining the dataframe content with specific styles
    '''
    tab_obj = viz_shared.reponsive_table_to_filtered_datapoints(data, customized_cols, header, exp_format)
    return tab_obj


def convert_relayout_data_to_df_reg(relayout_data, df, models):
    '''
    Convert raw data format from relayout selection range by user into the correct df fit for viz purpose

    Arguments:
        relayout_data (:obj:`Dict`):
            dictionary like data containing selection range indices returned from plotly graph
        df (:obj:`~pandas.DataFrame`):
            dataframe tap-out from interpreters pipeline
        models (:obj:`List[str]`):
            model names defined by user during spin-up of Tenjin app

    Returns:
        :obj:`~pandas.DataFrame`:
            dataframe fit for the responsive table-graph filtering
    '''
    if detected_single_xaxis(relayout_data):
        x_start_idx = int(relayout_data['xaxis.range[0]']) if relayout_data['xaxis.range[0]'] >= 0 else 0
        x_stop_idx = int(relayout_data['xaxis.range[1]']) if relayout_data['xaxis.range[1]'] <= len(df) - 1 else len(df) - 1
        df_filtered_x = df.iloc[df.index[x_start_idx]:df.index[x_stop_idx]]

        y_start_idx, y_stop_idx = get_min_max_offset(df_filtered_x, models)
        df_final = dataframe_prep_on_model_count_by_yaxis_slice(df_filtered_x, models, y_start_idx, y_stop_idx)

    elif detected_single_yaxis(relayout_data):
        y_start_idx = relayout_data['yaxis.range[0]']
        y_stop_idx = relayout_data['yaxis.range[1]']
        df_filtered_y = dataframe_prep_on_model_count_by_yaxis_slice(df, models, y_start_idx, y_stop_idx)

        x_start_idx, x_stop_idx = get_min_max_index(df_filtered_y, models, y_start_idx, y_stop_idx)
        x_start_idx = x_start_idx if x_start_idx >= 0 else 0
        x_stop_idx = x_stop_idx if x_stop_idx <= len(df_filtered_y) - 1 else len(df_filtered_y) - 1
        df_final = df_filtered_y.iloc[df_filtered_y.index[x_start_idx]:df_filtered_y.index[x_stop_idx]]

    else:  # a complete range is provided by user (with proper x-y coordinates)
        x_start_idx = int(relayout_data['xaxis.range[0]']) if relayout_data['xaxis.range[0]'] >= 0 else 0
        x_stop_idx = int(relayout_data['xaxis.range[1]']) if relayout_data['xaxis.range[1]'] <= len(df) - 1 else len(df) - 1
        y_start_idx = relayout_data['yaxis.range[0]']
        y_stop_idx = relayout_data['yaxis.range[1]']

        df_filtered = df.iloc[df.index[x_start_idx]:df.index[x_stop_idx]]
        df_final = dataframe_prep_on_model_count_by_yaxis_slice(df_filtered, models, y_start_idx, y_stop_idx)
    return df_final


def convert_relayout_data_to_df_cls(fig_class_label, relayout_data, df_feature, df_viz_specific):
    '''
    Convert raw data format from relayout selection range by user into the correct df fit for viz purpose

    Arguments:
        fig_class_label (str):
            class label name
        relayout_data (:obj:`Dict`):
            data containing selection range indices returned from plotly graph
        df (:obj:`~pandas.DataFrame`):
            dataframe tap-out from interpreters pipeline
        df_viz_specific (:obj:`~pandas.DataFrame`):
            dataframe prefiltered with right class label and model

    Returns:
        :obj:`~pandas.DataFrame`:
            dataframe fit for the responsive table-graph filtering
    '''
    relayout_dict = relayout_data[1]  # active relayout_data selected by user
    x_start_idx, x_stop_idx, y_start_idx, y_stop_idx = get_adjusted_xy_coordinate(relayout_dict, df_feature)

    lower_spec_limit_x = (df_viz_specific['index'] >= x_start_idx)
    upper_spec_limit_x = (df_viz_specific['index'] <= x_stop_idx)
    df_filtered = conditional_sliced_df(df_viz_specific, lower_spec_limit_x, upper_spec_limit_x)

    lower_spec_limit_y = (df_filtered[fig_class_label] >= y_start_idx)
    upper_spec_limit_y = (df_filtered[fig_class_label] <= y_stop_idx)
    df_final_prob = conditional_sliced_df(df_filtered, lower_spec_limit_y, upper_spec_limit_y)

    final_filtered_idx = list(df_final_prob.index)
    df_final_feature = df_feature[df_feature['index'].isin(final_filtered_idx)]
    return df_final_feature, df_final_prob


class MissPredictions:
    '''
    Main integration for feature component on Miss Prediction.

        - On Regression: To generate single miss-prediction scatter plot by data index points
        - On Classification: To generate scatter plots for probabilities comparison on correct data point vs miss-predicted data point \
            for each class label

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Important Attributes:

        - analysis_type (str):
            Analysis type defined by user during initial inputs preparation via data_loader stage.
        - model_names (:obj:`List[str]`):
            model names defined by user during initial inputs preparation via data_loader stage.
        - is_bimodal (bool):
            to indicate if analysis involves 2 models

    Returns:
        :obj:`~dash_core_components.Container`:
            styled dash components displaying graph and/or table objects
    '''
    def __init__(self, data_loader: Union[CSVDataLoader, DataframeLoader]):
        self.data_loader = data_loader
        self.analysis_type = data_loader.get_analysis_type()
        self.model_names = data_loader.get_model_list()
        self.is_bimodal = True if len(self.model_names) > 1 else False

        # instantiate here instead of under def show() as it will be used in callbacks as well
        if is_regression(self.analysis_type):
            self.preds_offset, self.df = fig_plot_prediction_offset_overview(self.data_loader)
            self.cols_table_reg = [col.replace('_', ' ') for col in self.df.columns]

        elif is_classification(self.analysis_type):
            compact_outputs = fig_probabilities_spread_pattern(self.data_loader)
            self.probs_pattern, self.label_state = compact_outputs[0], compact_outputs[1]
            self.dfs_viz, self.df_features, self.class_labels = compact_outputs[2], compact_outputs[3], compact_outputs[4]
            self.df_features = insert_index_col(self.df_features)
            self.dfs_viz = [insert_index_col(df) for df in self.dfs_viz]

    def show(self):
        '''
        Method to tapout styled html for misspredictions
        '''
        if is_regression(self.analysis_type):
            miss_preds = dbc.Container([
                                html.Div(html.H6(style_configs.INSTRUCTION_TEXT_SHARED), className='h6__dash-table-instruction-reg'),
                                dbc.Row(dcc.Graph(id='fig-reg',
                                                figure=self.preds_offset,),
                                                justify='center',
                                                className='border__common-misspred-reg'),
                                html.Div(id='alert-to-reset-reg'),
                                html.Div(id='show-feat-prob-table-reg', className='div__table-proba-misspred'),
                                html.Br()
                        ], fluid=True)

        elif is_classification(self.analysis_type):
            fig_objs_model_1 = self.probs_pattern[0]
            tables_model_1 = self.label_state[0]

            instruction_txt_shared = [html.Div(html.H6(style_configs.INSTRUCTION_TEXT_SHARED),
                                                        className='h6__dash-table-instruction-misspred-cls')]
            dash_table_ls_shared = [html.Div(id='main-title-plot-name'),
                                    html.Div(id='alert-to-reset-cls'),
                                    html.Div(id='table-title-misspred-features'),
                                    html.Div(id='show-feat-table', className='div__table-proba-misspred'),
                                    html.Br(),
                                    html.Div(id='table-title-misspred-probs'),
                                    html.Div(id='show-prob-table', className='div__table-proba-misspred'),
                                    html.Br()]

            if self.is_bimodal and is_classification(self.analysis_type):  # cover bimodal_binary and bimodal_multiclass
                fig_objs_model_2 = self.probs_pattern[1]
                tables_model_2 = self.label_state[1]

                dash_fig_ls = []
                for i in range(0, len(fig_objs_model_1), 2):
                    try:  # enabling the display of a pair of figures for better comparison view
                        fig_pair = dbc.Row([
                                        dbc.Col([
                                            dbc.Row([
                                                dbc.Col([
                                                    dbc.Row(
                                                        dcc.Graph(
                                                            id={'index': f'fig-cls-{self.model_names[0]}-labelcls-{self.class_labels[i]}',
                                                                'type': 'fig-obj-prob-spread'},
                                                            figure=fig_objs_model_1[i]),
                                                        justify='center'),
                                                    dbc.Row(
                                                        html.Div(
                                                            html.Div(tables_model_1[i], className='div__table-proba-misspred')),
                                                        justify='center')
                                                ]),
                                                dbc.Col([
                                                    dbc.Row(
                                                        dcc.Graph(
                                                            id={'index': f'fig-cls-{self.model_names[1]}-labelcls-{self.class_labels[i]}',
                                                                'type': 'fig-obj-prob-spread'},
                                                            figure=fig_objs_model_2[i]),
                                                        justify='center'),
                                                    dbc.Row(
                                                        html.Div(
                                                            html.Div(tables_model_2[i], className='div__table-proba-misspred')),
                                                        justify='center')
                                                ]),
                                            ])
                                        ], className='border__common-left'),

                                        dbc.Col([
                                            dbc.Row([
                                                dbc.Col([
                                                    dbc.Row(
                                                        dcc.Graph(
                                                            id={'index': f'fig-cls-{self.model_names[0]}-labelcls-{self.class_labels[i + 1]}',
                                                                'type': 'fig-obj-prob-spread'},
                                                            figure=fig_objs_model_1[i + 1]),
                                                        justify='center'),
                                                    dbc.Row(
                                                        html.Div(
                                                            html.Div(tables_model_1[i + 1], className='div__table-proba-misspred')),
                                                        justify='center')
                                                ]),
                                                dbc.Col([
                                                    dbc.Row(
                                                        dcc.Graph(
                                                            id={'index': f'fig-cls-{self.model_names[1]}-labelcls-{self.class_labels[i + 1]}',
                                                                'type': 'fig-obj-prob-spread'},
                                                            figure=fig_objs_model_2[i + 1]),
                                                        justify='center'),
                                                    dbc.Row(
                                                        html.Div(
                                                            html.Div(tables_model_2[i + 1], className='div__table-proba-misspred')),
                                                        justify='center')
                                                ]),
                                            ])
                                        ], className='border__common-right')
                                    ])
                        dash_fig_ls.append(fig_pair)

                    except IndexError:  # handling the last odd figure that can't be paired out
                        fig_pair = dbc.Row([
                                        dbc.Col([
                                            dbc.Row([
                                                dbc.Col([
                                                    dbc.Row(
                                                        dcc.Graph(
                                                            id={'index': f'fig-cls-{self.model_names[0]}-labelcls-{self.class_labels[i]}',
                                                                'type': 'fig-obj-prob-spread'},
                                                            figure=fig_objs_model_1[i]),
                                                        justify='center'),
                                                    dbc.Row(
                                                        html.Div(
                                                            html.Div(tables_model_1[i], className='div__table-proba-misspred')),
                                                        justify='center')
                                                ]),
                                                dbc.Col([
                                                    dbc.Row(
                                                        dcc.Graph(
                                                            id={'index': f'fig-cls-{self.model_names[1]}-labelcls-{self.class_labels[i]}',
                                                                'type': 'fig-obj-prob-spread'},
                                                            figure=fig_objs_model_2[i]),
                                                        justify='center'),
                                                    dbc.Row(
                                                        html.Div(
                                                            html.Div(tables_model_2[i], className='div__table-proba-misspred')),
                                                        justify='center')
                                                ])
                                            ])
                                        ], className='border__common')])
                        dash_fig_ls.append(fig_pair)

                compiled_fig_table_objs = instruction_txt_shared + dash_fig_ls + dash_table_ls_shared
                miss_preds = dbc.Container(compiled_fig_table_objs, fluid=True)

            elif not self.is_bimodal and 'binary' in self.analysis_type:  # single modal binary classification
                dash_fig_ls = dbc.Row([
                                    dbc.Col([
                                        dbc.Row(
                                            dcc.Graph(
                                                id={'index': f'fig-cls-{self.model_names[0]}-labelcls-{self.class_labels[0]}',
                                                    'type': 'fig-obj-prob-spread'},
                                                figure=fig_objs_model_1[0]),
                                            justify='center'),
                                        dbc.Row(html.Div(
                                            html.Div(tables_model_1[0], className='div__table-proba-misspred')), justify='center'),
                                    ]),
                                    dbc.Col([
                                        dbc.Row(
                                            dcc.Graph(
                                                id={'index': f'fig-cls-{self.model_names[0]}-labelcls-{self.class_labels[1]}',
                                                    'type': 'fig-obj-prob-spread'},
                                                figure=fig_objs_model_1[1]),
                                            justify='center'),
                                        dbc.Row(
                                            html.Div(
                                                html.Div(tables_model_1[1], className='div__table-proba-misspred')),
                                            justify='center'),
                                    ])], className='border__common-misspred-cls-single-binary')

                instruction_txt_cls_single_binary = [html.Div(
                                                        html.H6(style_configs.INSTRUCTION_TEXT_SHARED),
                                                        className='h6__dash-table-instruction-cls-single-binary')]

                compiled_fig_table_objs = instruction_txt_cls_single_binary + [dash_fig_ls] + dash_table_ls_shared
                miss_preds = dbc.Container(compiled_fig_table_objs, fluid=True)

            elif not self.is_bimodal and 'multiclass' in self.analysis_type:  # single modal multi-class classification
                dash_fig_ls = []
                for i in range(0, len(fig_objs_model_1)):
                    fig_pair = dbc.Col([
                                    dbc.Row([dcc.Graph(id={'index': f'fig-cls-{self.model_names[0]}-labelcls-{self.class_labels[i]}',
                                                        'type': 'fig-obj-prob-spread'},
                                                    figure=fig_objs_model_1[i])], justify='center'),
                                    dbc.Row(
                                        html.Div(
                                            html.Div(tables_model_1[i], className='div__table-proba-misspred')),
                                        justify='center'),
                                ], className='border__common')
                    dash_fig_ls.append(fig_pair)

                compiled_fig_table_objs = instruction_txt_shared + [dbc.Row(dash_fig_ls)] + dash_table_ls_shared
                miss_preds = dbc.Container(compiled_fig_table_objs, fluid=True)
        return miss_preds

    def callbacks(self):
        @app.callback(
            Output('alert-to-reset-reg', 'children'),
            Output('show-feat-prob-table-reg', 'children'),
            Input('fig-reg', 'relayoutData'),  # taking in range selection info from figures
            Input('fig-reg', 'restyleData'))  # taking in legend filtration info from figures
        def display_relayout_data_reg(relayout_data, restyle_data):
            '''
            Callbacks functionalities for regression tasks
            '''
            if relayout_data is not None:
                try:
                    self.df = self.df.round(2)  # to limit table cell having values with long decimals for better viz purpose
                except TypeError:
                    self.df

                if is_reset(relayout_data):
                    collapsed_header = style_configs.collapse_header_style()

                    table_obj_reg = table_with_relayout_datapoints('', self.cols_table_reg, collapsed_header, 'none')
                    alert_obj_reg = style_configs.dummy_alert()

                elif is_active_trace(relayout_data):
                    models = self.model_names
                    if restyle_data is not None:  # [{'visible': ['legendonly']}, [1]]
                        if detected_legend_filtration(restyle_data):
                            model_to_exclude_from_view = self.model_names[restyle_data[1][0]]
                            models = [model for model in self.model_names if model != model_to_exclude_from_view]

                    df_final = convert_relayout_data_to_df_reg(relayout_data, self.df, models)
                    df_final.columns = self.cols_table_reg  # to have customized column names displayed on table

                    default_header = style_configs.default_header_style()
                    data_relayout_reg = df_final.to_dict('records')
                    table_obj_reg = table_with_relayout_datapoints(data_relayout_reg, self.cols_table_reg, default_header, 'csv')
                    alert_obj_reg = style_configs.activate_alert()
                return alert_obj_reg, table_obj_reg
            else:
                raise PreventUpdate

        @app.callback(
            Output('alert-to-reset-cls', 'children'),
            Output('main-title-plot-name', 'children'),
            Output('table-title-misspred-features', 'children'),
            Output('show-feat-table', 'children'),
            Output('table-title-misspred-probs', 'children'),
            Output('show-prob-table', 'children'),
            Input({'index': ALL, 'type': 'fig-obj-prob-spread'}, 'relayoutData'),
            Input({'index': ALL, 'type': 'fig-obj-prob-spread'}, 'restyleData'))
        def display_relayout_data_cls(relayout_data, restyle_data):
            '''
            Callbacks functionalities for classification tasks
            '''
            fig_obj_ids = [fig_id_dict['id']['index'] for fig_id_dict in dash.callback_context.inputs_list[0]]

            default_plot_name = style_configs.DEFAULT_PLOT_NAME_STYLE
            default_title = style_configs.DEFAULT_TITLE_STYLE
            title_main_plot_name = html.H5('', style=default_plot_name, className='title__main-plot-name-cls')
            title_table_features = html.H6('Feature Values :', style=default_title, className='title__table-misspred-cls')
            title_table_probs = html.H6('Probabilities Overview :', style=default_title, className='title__table-misspred-cls')

            if not all(item is None for item in relayout_data):
                try:
                    self.df_features = self.df_features.round(2)  # limit long decimals on feature values
                    self.dfs_viz[0] = self.dfs_viz[0].round(4)  # standardize prob values to 4 decimals
                    if self.is_bimodal:
                        self.dfs_viz[1] = self.dfs_viz[1].round(4)
                except TypeError:
                    self.df_features
                    self.dfs_viz

                models_ref_dict = {model: i for i, model in enumerate(self.model_names)}

                collapsed_header = style_configs.collapse_header_style()
                default_title = style_configs.hidden_title_style()
                default_plot_name = style_configs.hidden_plot_name_style()

                table_obj_cls_features = table_with_relayout_datapoints('', list(self.df_features.columns), collapsed_header, 'none')
                table_obj_cls_probs = table_with_relayout_datapoints('', list(self.dfs_viz[0].columns), collapsed_header, 'none')

                if detected_unique_figure(relayout_data):  # unique fig trace (only one fig is used to generate info table)
                    specific_relayout = identify_active_trace(relayout_data)  # output as tuple (idx, relayout_data)

                    fig_id = fig_obj_ids[specific_relayout[0]]
                    fig_class_label = fig_id.split('-cls-')[-1].split('-labelcls-')[-1]
                    model_in_view = fig_id.split('-cls-')[-1].split('-labelcls-')[0]
                    title_main_plot_name = html.H5(f'Currently inspecting : class {fig_class_label} [ {model_in_view} ]',
                                                    style=default_plot_name,
                                                    className='title__main-plot-name-cls')

                    specific_df_viz_id = models_ref_dict[model_in_view]
                    df_viz_model_in_view = self.dfs_viz[specific_df_viz_id]
                    df_viz_specific = df_viz_model_in_view[df_viz_model_in_view['yTrue'].astype('str') == fig_class_label]

                    specific_restyle_data = restyle_data[specific_relayout[0]]
                    if specific_restyle_data is not None:
                        if detected_legend_filtration(specific_restyle_data):
                            if specific_restyle_data[1][0] == 1:
                                data_field_to_exclude = 'miss-predict'
                            else:
                                data_field_to_exclude = 'correct'
                            df_viz_specific = df_viz_specific[df_viz_specific['pred_state'] != data_field_to_exclude]
                        else:
                            df_viz_specific

                    df_filtered_feature, df_filtered_viz = convert_relayout_data_to_df_cls(fig_class_label,
                                                                                            specific_relayout,
                                                                                            self.df_features,
                                                                                            df_viz_specific)

                    df_filtered_viz.columns = [col.replace('_', ' ') for col in df_filtered_viz.columns]
                    data_relayout_features = df_filtered_feature.to_dict('records')

                    # activate visibility to prepare rendering of data table upon completion of range selection and/or legend filtration
                    default_header = style_configs.default_header_style()
                    default_title['visibility'] = 'visible'
                    default_plot_name['visibility'] = 'visible'

                    title_main_plot_name = html.H5(f'Currently inspecting : class {fig_class_label} [ {model_in_view} ]',
                                                    style=default_plot_name,
                                                    className='title__main-plot-name-cls')
                    table_obj_cls_features = table_with_relayout_datapoints(data_relayout_features,
                                                                            list(self.df_features.columns),
                                                                            default_header,
                                                                            'csv')

                    data_relayout_probs = df_filtered_viz.to_dict('records')
                    table_obj_cls_probs = table_with_relayout_datapoints(data_relayout_probs,
                                                                        list(df_filtered_viz.columns),
                                                                        default_header,
                                                                        'csv')
                    alert_obj_cls = dbc.Alert(color="light")
                elif detected_more_than_1_unique_figure(relayout_data):
                    alert_obj_cls = dbc.Alert(style_configs.WARNING_TEXT, color="danger", style={'text-align': 'center'})
                else:  # user reset data range
                    alert_obj_cls = dbc.Alert(color="light")

                return alert_obj_cls, title_main_plot_name, title_table_features, table_obj_cls_features, title_table_probs, table_obj_cls_probs

            else:
                raise PreventUpdate



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import dash_bootstrap_components as dbc
import dash_html_components as html


INSTRUCTION_TEXT_SHARED = 'Click and drag on the graph to select the range of data points to inspect feature values.'
INSTRUCTION_TEXT_REG = 'To reset back to default settings, hover over icons on the top right of the graph and click "Reset axes" icon.'
WARNING_TEXT = 'To inspect new range of datapoints in different graph, please first reset the earlier selection by clicking "Reset axes" icon ' \
                'at the top right corner of the graph.'

DEFAULT_HEADER_STYLE = {'fontWeight': 'bold', 'color': 'white', 'backgroundColor': '#7e746d', 'border': '1px solid white', 'height': '45px'}
DEFAULT_TITLE_STYLE = {'visibility': 'visible'}
DEFAULT_PLOT_NAME_STYLE = {'visibility': 'visible'}

DEFAULT_RANGE_SELECTION_TEXT_REG = "Enter range of data to compare distribution ( default slicing - last 20% of dataset ) :"
DEFAULT_RANGE_SELECTION_TEXT_CLS = "Enter range of data to compare distribution ( default slicing - full range ):"

DEFAULT_SLIDER_RANGE = {n: str(n) for n in range(1, 11)}

OPTIONS_NO_OF_CLUSTERS = [{'label': f'{n}', 'value': f'{n}'} for n in range(2, 9)]  # option 2 to 8
LOG_METHOD_DICT = {'log': math.log, 'log1p': math.log1p, 'log2': math.log2, 'log10': math.log10}


def default_header_style():
    DEFAULT_HEADER_STYLE['visibility'] = 'visible'
    return DEFAULT_HEADER_STYLE


def hidden_title_style():
    DEFAULT_TITLE_STYLE['visibility'] = 'hidden'
    return DEFAULT_TITLE_STYLE


def hidden_plot_name_style():
    DEFAULT_PLOT_NAME_STYLE['visibility'] = 'hidden'
    return DEFAULT_PLOT_NAME_STYLE


def collapse_header_style():
    COLLAPSE_HEADER_STYLE = DEFAULT_HEADER_STYLE.copy()
    COLLAPSE_HEADER_STYLE['border'] = 'none'
    COLLAPSE_HEADER_STYLE['visibility'] = 'collapse'
    return COLLAPSE_HEADER_STYLE


def input_range_subnote(df):
    return f'Available dataset range : from 0 to max at {len(df)}'


def dummy_alert():
    alert_obj = dbc.Alert(color="light", style={'visibility': 'hidden'})
    return alert_obj


def activate_alert():
    alert_obj = dbc.Alert(INSTRUCTION_TEXT_REG, color="secondary", className='alert__note-reg')
    return alert_obj


def activate_cluster_error_alert(label_class):
    err_message = f'Miss prediction data points are not sufficient for auto-clustering in < Class {label_class} >. ' \
                    'Minimum number of datapoint for auto-clustering is 8 datapoints per class per model'
    alert_obj = dbc.Alert(err_message, color='warning', dismissable=True, is_open=True)
    return alert_obj


def activate_range_input_error_alert():
    err_message = 'Invalid data range detected. ' \
                    'Please input a valid data range format ' \
                    '(  example:   start_idx:stop_idx   =>   200:1000   or   25%:75%    with   start_idx  <  stop_idx  )'
    alert_obj = dbc.Alert(html.Pre(err_message, className='html_Pre__alert-message'), color='warning', dismissable=True, is_open=True)
    return alert_obj


def activate_invalid_limit_alert(df):
    err_message = f'Invalid data range detected. Allowable index range for slicing :   0 to {len(df)}   or   0% to 100%'
    alert_obj = dbc.Alert(html.Pre(err_message, className='html_Pre__alert-message'), color='warning', dismissable=True, is_open=True)
    return alert_obj


def activate_incomplete_range_entry_alert():
    err_message = 'Invalid data range detected. Please enter a complete range ' \
                    '(  example:   start_idx:stop_idx   =>   200:1000   or   25%:75%    with   start_idx  <  stop_idx  )'
    alert_obj = dbc.Alert(html.Pre(err_message, className='html_Pre__alert-message'), color='warning', dismissable=True, is_open=True)
    return alert_obj


def activate_invalid_index_input_alert(df):
    err_message = f'Invalid data index detected. Allowable index range :   0 to {len(df)}   ' \
                    'separate index with  ","  if want to inspect more than 1 index ( example: 1, 23, 456 )'
    alert_obj = dbc.Alert(html.Pre(err_message, className='html_Pre__alert-message'), color='warning', dismissable=True, is_open=True)
    return alert_obj


def no_error_alert():
    alert_obj = dbc.Alert('', color='warning', dismissable=True, is_open=False)
    return alert_obj



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List
import pandas as pd

import dash
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
import dash_core_components as dcc
import dash_html_components as html
import dash_bootstrap_components as dbc

from rarity.app import app
from rarity.data_loader import CSVDataLoader, DataframeLoader
from rarity.interpreters.structured_data import IntSimilaritiesCounterFactuals
from rarity.visualizers import shared_viz_component as viz_shared
from rarity.utils import style_configs
from rarity.utils.common_functions import is_regression, is_classification, get_max_value_on_slider, \
                                            detected_bimodal, detected_invalid_index_inputs


def generate_similarities(data_loader: Union[CSVDataLoader, DataframeLoader], user_defined_idx, feature_to_exclude=None, top_n=3):
    '''
    Tapout table collating feature info corresponding to user defined index and top N index based on distance score.
    Applicable to both regression and classification

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module
        user_defined_idx (int):
            Index of the data point of interest specified by user
        feature_to_exclude (:obj:`List[str]`, `optional`):
                A list of features to be excluded from the ranking and similarities distance calculation
        top_n (int):
                Number indicating the max limit of records to be displayed based on the distance ranking

    Returns:
        :obj:`~dash_table.DataTable`:
            table object outlining the dataframe content with dynamic-conditional styles
    '''
    df_viz, idx_for_top_n, calculated_distance = IntSimilaritiesCounterFactuals(data_loader).xform(user_defined_idx, feature_to_exclude, top_n)
    df_top_n = _base_df_by_calculated_distance(df_viz, idx_for_top_n, calculated_distance)

    category = ['User_defined_idx']
    category_top_n = [f'Top_{i + 1}' for i in range(len(df_top_n) - 1)]  # len(df_viz) - 1 as first row is user_defined_idx
    category = category + category_top_n
    df_top_n.insert(0, 'category', category)

    feature_cols = list(data_loader.get_features().columns)
    if is_regression(data_loader.get_analysis_type()):
        try:
            # to limit table cell having values with long decimals for better viz purpose
            df_top_n[list(df_top_n.columns)[3:]] = df_top_n[list(df_top_n.columns)[3:]].round(2)
        except TypeError:
            df_top_n
    elif is_classification(data_loader.get_analysis_type()):
        for col in df_top_n.columns:
            if 'float' in str(df_top_n[col].dtypes):
                df_top_n[col] = df_top_n[col].apply(lambda x: round(x, 4))

    table_obj = viz_shared.reponsive_table_to_filtered_datapoints_similaritiesCF(df_top_n,
                                                                                list(df_top_n.columns),
                                                                                feature_cols,
                                                                                style_configs.DEFAULT_HEADER_STYLE,
                                                                                'csv')
    return table_obj


def generate_counterfactuals(data_loader: Union[CSVDataLoader, DataframeLoader], user_defined_idx, feature_to_exclude=None, top_n=3):
    '''
    Tapout table collating feature info corresponding to user defined index and top N index based on distance score with condition
    that the prediction labels of top N index differ from prediction label of user defined index
    Applicable to both classification only

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module
        user_defined_idx (int):
            Index of the data point of interest specified by user
        feature_to_exclude (:obj:`List[str]`, `optional`):
                A list of features to be excluded from the ranking and similarities distance calculation
        top_n (int):
                Number indicating the max limit of records to be displayed based on the distance ranking

    Returns:
        :obj:`~dash_table.DataTable`:
            table object outlining the dataframe content with dynamic-conditional styles
    '''
    org_data_size = len(data_loader.get_features())
    df_viz, idx_sorted_by_distance, calculated_distance = IntSimilaritiesCounterFactuals(data_loader).xform(user_defined_idx,
                                                                                                            feature_to_exclude,
                                                                                                            org_data_size)
    df_top_n = _base_df_by_calculated_distance(df_viz, idx_sorted_by_distance, calculated_distance)

    table_objs = []
    for model in data_loader.get_model_list():
        usr_pred_label = df_top_n.loc[lambda x: x['index'] == user_defined_idx, f'yPred_{model}'].values[0]
        usr_true_label = df_top_n.loc[lambda x: x['index'] == user_defined_idx, 'yTrue'].values[0]

        df_user_idx = df_top_n.loc[lambda x: x['index'] == user_defined_idx, :]
        df_filtered_cf = df_top_n[(df_top_n['yTrue'] == usr_true_label) & (df_top_n[f'yPred_{model}'] != usr_pred_label)].head(top_n)
        df_top_n_cf = pd.concat([df_user_idx, df_filtered_cf], axis=0)

        category = ['User_defined_idx']
        category_top_n = [f'Top_CounterFactual_{i + 1}' for i in range(len(df_top_n_cf) - 1)]  # len(df_top_n_cf)-1 : 1st row is user_defined_idx
        category = category + category_top_n
        df_top_n_cf.insert(0, 'category', category)

        feature_cols = list(data_loader.get_features().columns)
        for col in df_top_n_cf.columns:
            if 'float' in str(df_top_n_cf[col].dtypes):
                df_top_n_cf[col] = df_top_n_cf[col].apply(lambda x: round(x, 4))

        table_obj = viz_shared.reponsive_table_to_filtered_datapoints_similaritiesCF(df_top_n_cf,
                                                                                    list(df_top_n_cf.columns),
                                                                                    feature_cols,
                                                                                    style_configs.DEFAULT_HEADER_STYLE,
                                                                                    'csv')
        table_objs.append(table_obj)
    return table_objs


def _base_df_by_calculated_distance(df: pd.DataFrame, idx_sorted_by_distance: List[int], calculated_distance: List[float]):
    '''
    Setup new dataframe storing calculated distance info
    '''
    df_top_n = pd.DataFrame()
    df_top_n['index'] = idx_sorted_by_distance
    df_top_n['calculated_distance'] = calculated_distance
    df_top_n = df_top_n.merge(df, how='left', on='index')
    return df_top_n


def _table_objs_similarities(data_loader: Union[CSVDataLoader, DataframeLoader],
                            user_defined_idx: int,
                            feature_to_exclude: List[str],
                            top_n: int):
    '''
    List collating layouts for similarities table based on user index/indices
    '''
    table_objs_similarities = []
    for idx in str(user_defined_idx).replace(' ', '').split(','):
        table_obj_similarities = generate_similarities(data_loader, int(idx), feature_to_exclude, top_n)
        row_layout_single_idx_table_obj = dbc.Row(
                                                html.Div(table_obj_similarities,
                                                        className='div__table-proba-misspred'),
                                                justify='center')
        table_objs_similarities.append(row_layout_single_idx_table_obj)
    return table_objs_similarities


def _table_objs_counterfactuals(data_loader: Union[CSVDataLoader, DataframeLoader],
                                user_defined_idx: int,
                                feature_to_exclude: List[str],
                                top_n: int):
    '''
    List collating layouts for counterfactual tables based on user index/indices
    '''
    models = data_loader.get_model_list()
    table_objs_counterfactuals = []
    for idx in str(user_defined_idx).replace(' ', '').split(','):
        table_objs_cf_single_idx = generate_counterfactuals(data_loader, int(idx), feature_to_exclude, top_n)
        table_objs_counterfactuals.append(table_objs_cf_single_idx)

    viz_table_objs_counterfactuals = []
    for table_objs_cf in table_objs_counterfactuals:
        if detected_bimodal(models):
            cf_table_obj_single_idx = [dbc.Row([html.H5('Counter-Factuals for model : ',
                                                        id='title-after-topn-bottomn-reg',
                                                        className='h5__counterfactuals-section-title'),
                                                        html.H5(f'{models[0]}', className='h5__counterfactual-model')]),
                                                dbc.Row(
                                                    html.Div(table_objs_cf[0],
                                                            id='table-obj-similarities-bm-1',
                                                            className='div__table-proba-misspred'),
                                                    justify='center'),
                                                html.Br(),
                                                dbc.Row([dbc.Row(html.H5('Counter-Factuals for model : ',
                                                                id='title-after-topn-bottomn-reg',
                                                                className='h5__counterfactuals-section-title')),
                                                        dbc.Row(html.H5(f'{models[1]}', className='h5__counterfactual-model'))]),
                                                dbc.Row(
                                                    html.Div(table_objs_cf[1],
                                                            id='table-obj-similarities-bm-2',
                                                            className='div__table-proba-misspred'),
                                                    justify='center')]
            viz_table_objs_counterfactuals += cf_table_obj_single_idx
        else:
            cf_table_obj_single_idx = [dbc.Row([html.H5('Counter-Factuals for model : ',
                                                        id='title-after-topn-bottomn-reg',
                                                        className='h5__counterfactuals-section-title'),
                                                        html.H5(f'{models[0]}', className='h5__counterfactual-model')]),
                                                dbc.Row(
                                                    html.Div(table_objs_cf[0],
                                                            id='table-obj-similarities-sm',
                                                            className='div__table-proba-misspred'),
                                                    justify='center')]
            viz_table_objs_counterfactuals += cf_table_obj_single_idx
    return viz_table_objs_counterfactuals


class SimilaritiesCF:
    '''
    Main integration for feature component on Similarities-CounterFactuals

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Important Attributes:

        analysis_type (str):
            Analysis type defined by user during initial inputs preparation via data_loader stage.
        df_features (:obj:`~pandas.DataFrame`):
            Dataframe storing all features used in dataset
        feature_to_exclude (:obj:`List[str]`, `optional`):
            A list of features to be excluded from the ranking and similarities distance calculation
        user_defined_idx (int):
            Index of the data point of interest specified by user
        top_n (int):
                Number indicating the max limit of records to be displayed based on the distance ranking

    Returns:
        :obj:`~dash_core_components.Container`:
            styled dash components displaying graph and/or table objects
    '''
    def __init__(self, data_loader: Union[CSVDataLoader, DataframeLoader]):
        self.data_loader = data_loader
        self.analysis_type = data_loader.get_analysis_type()
        self.df_features = data_loader.get_features()
        self.feature_to_exclude = []
        self.user_defined_idx = '1'
        self.top_n = 3
        self.table_objs_similarities = _table_objs_similarities(self.data_loader, self.user_defined_idx, self.feature_to_exclude, self.top_n)

    def show(self):
        '''
        Method to tapout styled html for Similarities-CounterFactuals
        '''
        options_feature_ls = [{'label': f'{col}', 'value': f'{col}'} for col in self.df_features.columns]
        shared_layout_reg_cls = [
                                    html.Div([
                                        dbc.Row([
                                            dbc.Col([
                                                dbc.Row(html.Div(html.H6('Specify data index ( default index: 1 ) :'),
                                                                        className='h6__similaritiesCF-index')),
                                                dbc.Row(dbc.Input(id='input-specific-idx-similaritiesCF',
                                                                    placeholder='example:  1   OR   12, 123, 1234',
                                                                    type='text',
                                                                    value=None)),
                                                dbc.Row(html.Div(html.Pre(style_configs.input_range_subnote(self.df_features),
                                                                            className='text__range-header-kldiv-featfist')))], width=6),
                                            dbc.Col([
                                                dbc.Row(html.Div(html.H6('Select feature to exclude from similarities calculation '
                                                                        '( if applicable ) :'),
                                                                        className='h6__feature-to-exclude')),
                                                dbc.Row(dbc.Col(dcc.Dropdown(id='select-feature-to-exclude-similaritiesCF', 
                                                                            options=options_feature_ls,
                                                                            value=[], multi=True)))], width=6)]),
                                        html.Br(),
                                        dbc.Row([
                                            dbc.Col([
                                                dbc.Row(html.Div(html.H6('Select number of records to display'),
                                                                            className='h6__display-top-bottom-n')),
                                                dbc.Row(html.Div(html.Span('( ranked by calculated distance on overall feature similarites '
                                                                            'referencing to the index defined above ):'),
                                                                className='text__display-header-similaritiesCF')),
                                                dcc.Slider(id='select-slider-top-n-similaritiesCF',
                                                    min=1,
                                                    max=get_max_value_on_slider(self.df_features, 'similaritiesCF'),  # max at 10
                                                    step=1,
                                                    value=3,
                                                    marks=style_configs.DEFAULT_SLIDER_RANGE)], width=10),
                                            dbc.Col(dbc.Row(
                                                        dcc.Loading(id='loading-output-similaritiesCF',
                                                            type='circle', color='#a80202'),
                                                    justify='left', className='loading__similaritiesCF'), width=1),
                                            dbc.Col(dbc.Row(
                                                        dbc.Button("Update",
                                                                    id='button-similaritiesCF-update',
                                                                    n_clicks=0,
                                                                    className='button__update-dataset'), justify='right'))])
                                    ], className='border__select-dataset'),
                                    html.Div(id='alert-index-input-error-similaritiesCF'),
                                    html.Br(),
                                    dbc.Row(html.H5('Comparison based on Feature Similarities',
                                                    id='title-after-topn-bottomn-reg',
                                                    className='h5__counterfactuals-section-title'))
                                ] + [html.Div(self.table_objs_similarities, id='table-objs-similarities-shared')]

        if is_regression(self.analysis_type):
            similaritiesCF = dbc.Container(shared_layout_reg_cls + [html.Div(id='table-objs-counter-factuals')], fluid=True)

        elif is_classification(self.analysis_type):
            self.models = self.data_loader.get_model_list()
            self.table_objs_counterfactuals = _table_objs_counterfactuals(self.data_loader,
                                                                            self.user_defined_idx,
                                                                            self.feature_to_exclude,
                                                                            self.top_n)

            combined_layouts_similaritiesCF = shared_layout_reg_cls + [html.Br(),
                                                                        html.Div(self.table_objs_counterfactuals,
                                                                                id='table-objs-counter-factuals')]
            similaritiesCF = dbc.Container(combined_layouts_similaritiesCF, fluid=True)
        return similaritiesCF

    def callbacks(self):
        @app.callback(
            Output('loading-output-similaritiesCF', 'children'),
            Output('alert-index-input-error-similaritiesCF', 'children'),
            Output('table-objs-similarities-shared', 'children'),
            Output('table-objs-counter-factuals', 'children'),
            Input('button-similaritiesCF-update', 'n_clicks'),
            State('input-specific-idx-similaritiesCF', 'value'),
            State('select-feature-to-exclude-similaritiesCF', 'value'),
            State('select-slider-top-n-similaritiesCF', 'value'))
        def generate_table_objs_based_on_user_selected_params(click_count, specific_idx, feature_to_exclude, top_n):
            '''
            Callbacks functionalities on params related to top-n, botton-n / both on regression and classification tasks
            '''
            if click_count > 0:
                if specific_idx is None:  # during first spin-up
                    specific_idx = self.user_defined_idx  # default value
                idx_input_err_alert = style_configs.no_error_alert()

                if detected_invalid_index_inputs(specific_idx, self.df_features):
                    idx_input_err_alert = style_configs.activate_invalid_index_input_alert(self.df_features)
                    return '', idx_input_err_alert, dash.no_update, dash.no_update

                if is_regression(self.analysis_type):
                    table_objs_similarities = _table_objs_similarities(self.data_loader, specific_idx, feature_to_exclude, top_n)
                    return '', idx_input_err_alert, table_objs_similarities, ''
                elif is_classification(self.analysis_type):
                    table_objs_similarities = _table_objs_similarities(self.data_loader, specific_idx, feature_to_exclude, top_n)
                    table_objs_counterfactuals = _table_objs_counterfactuals(self.data_loader, specific_idx, feature_to_exclude, top_n)
                    return '', idx_input_err_alert, table_objs_similarities, table_objs_counterfactuals
            else:
                raise PreventUpdate



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List
import pandas as pd
import numpy as np
from sklearn import metrics

from rarity.data_loader import CSVDataLoader, DataframeLoader
from rarity.utils.common_functions import is_regression, is_classification

ERR_DESC = ['Mean Absolute Error', 'Mean Squared Error', 'Root Mean Squared Error', 'R Squared']
ERR_NAMES = ['MAE', 'MSE', 'RMSE', 'R2']
METRICS_FORMULA = ['![MAE](assets/MAE.png)', '![MSE](assets/MSE.png)', 
                    '![RMSE](assets/RMSE.png)', '![R2](assets/R2.png)']
ERR_METRICS_DICT = {}
ERR_METRICS_DICT['Metrics Description'] = ERR_DESC
ERR_METRICS_DICT['Formula'] = METRICS_FORMULA
ERR_METRICS_DICT['Metrics Name'] = ERR_NAMES


class IntGeneralMetrics:
    """
    Transform raw data into input format suitable for visualization. General metrics cover
    confusion matrix, classification report, roc curve and precisionRecall curve

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

        viz_plot (str):
            Supported visualization types : ``confMat``, ``classRpt``, ``rocAUC``, ``preRacall``, ``stdErr``, None

    Important Attributes:

        - analysis_type (str):
            Analysis type defined by user during initial inputs preparation via data_loader stage.

    Returns:
        :obj:`~pd.DataFrame`:
            Dataframe with essential info suitable for visualization on regression task

    .. note::

            if classification, returns:

            - yTrue data in :obj:`~pd.Series`
            - yPred data in :obj:`~pd.Series` for [``confMat``, ``classRpt``] or :obj:`~pd.Dataframe` for [``rocAuc``, ``precRecall``]

            If multiclass, returns:

            - yPred data in :obj:`List[List[Tuple]]` pairing class label and yPred in :obj:`~pd.Series`
            - model_names in :obj:`List[str]`
    """
    def __init__(self, data_loader: Union[CSVDataLoader, DataframeLoader], viz_plot: str = None):
        self.data_loader = data_loader
        self.viz_plot = viz_plot
        self.analysis_type = self.data_loader.get_analysis_type()

    def _std_err_metrics(self, yTrue: pd.Series, yPred: pd.Series) -> List:
        mae = metrics.mean_absolute_error(yTrue, yPred)
        mse = metrics.mean_squared_error(yTrue, yPred)
        rmse = np.sqrt(metrics.mean_squared_error(yTrue, yPred))
        r2_score = metrics.r2_score(yTrue, yPred)
        return [int(mae), int(mse), int(rmse), round(r2_score, 4)]

    def xform(self):
        '''
        Core transformation function to tap-out data into input format suitable for plotly graph
        '''
        if is_regression(self.analysis_type):
            if self.viz_plot == 'stdErr':
                yTrue = self.data_loader.get_yTrue()
                yPreds = self.data_loader.get_yPreds()
                models = self.data_loader.get_model_list()
                if len(models) == 1:
                    ERR_METRICS_DICT[f'Model_{models[0]}'] = self._std_err_metrics(yTrue, yPreds)
                elif len(models) > 1:
                    for i in range(len(models)):
                        ERR_METRICS_DICT[f'Model_{models[i]}'] = self._std_err_metrics(yTrue, yPreds[yPreds.columns[i]])
                df = pd.DataFrame(ERR_METRICS_DICT)
            else:
                df = self.data_loader.get_all()
            return df

        elif is_classification(self.analysis_type):
            model_names = self.data_loader.get_model_list()
            yTrue = self.data_loader.get_yTrue()
            yTrue = yTrue['yTrue'].astype('string')
            preds = self.data_loader.get_yPreds()

            if self.viz_plot in ['confMat', 'classRpt']:
                yPred = [pred['yPred-label'] for pred in preds]
            elif self.viz_plot in ['rocAuc', 'precRecall']:
                yPred = [pred[pred.columns[-3]] for pred in preds]
                is_multiclass = len(set(yTrue)) > 2
                if is_multiclass:
                    yPred = []
                    for pred in preds:
                        label_keys = pred['yPred-label']
                        pred_values = pred[pred.columns[:-2]].max(axis=1)
                        pred_tmp = [(label_keys[i], pred_values[i]) for i in range(len(label_keys))]
                        yPred.append(pred_tmp)
            return yTrue, yPred, model_names



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List
import string
import pandas as pd


def identify_active_trace(relayout_data: Dict):
    '''
    only 1 unique trace should be in relayout_data
    this function is to retrieve the position info of the unique trace
    '''
    for idx, rdata in enumerate(relayout_data):
        if rdata is not None:
            if is_active_trace(rdata):
                specific_layout = idx, rdata
        else:
            continue
    return specific_layout


def is_active_trace(relayout_data: Dict):
    if ('xaxis.range[0]' in relayout_data.keys() or 'yaxis.range[0]' in relayout_data.keys()) and 'xaxis.autorange' not in relayout_data.keys():
        return True
    else:
        return False


def is_reset(relayout_data: Dict):
    if 'xaxis.autorange' in relayout_data.keys():
        return True
    else:
        return False


def is_regression(analysis_type: str):
    if analysis_type == 'regression':
        return True
    else:
        return False


def is_classification(analysis_type: str):
    if 'classification' in analysis_type:
        return True
    else:
        return False


def detected_bimodal(models: List[str]):
    if len(models) == 2:
        return True
    else:
        return False


def detected_invalid_index_inputs(specific_idx: str, df: pd.DataFrame):
    special_characters = list(string.punctuation.replace(',', ''))
    letters = list(string.ascii_letters)
    if any(sp_char for sp_char in special_characters if sp_char in specific_idx):
        return True
    elif any(letter for letter in letters if letter in specific_idx):
        return True
    elif any(idx for idx in specific_idx.replace(' ', '').split(',') if (int(idx) > len(df)) or (int(idx) < 0)):
        return True


def detected_legend_filtration(restyle_data: Dict):
    if restyle_data[0]['visible'] == ['legendonly']:
        return True
    elif restyle_data[0]['visible'] == [True]:
        return False
    else:
        return False


def detected_unique_figure(relayout_data: Dict):
    if sum(_valid_fig_key_ls(relayout_data)) == 1:
        return True
    else:
        return False


def detected_more_than_1_unique_figure(relayout_data: Dict):
    if sum(_valid_fig_key_ls(relayout_data)) > 1:
        return True
    else:
        return False


def detected_single_xaxis(relayout_data: Dict):
    if 'yaxis.range[0]' not in relayout_data.keys():
        return True
    else:
        return False


def detected_single_yaxis(relayout_data: Dict):
    if 'xaxis.range[0]' not in relayout_data.keys():
        return True
    else:
        return False


def detected_complete_pair_xaxis_yaxis(relayout_data: Dict):
    if 'xaxis.range[0]' in relayout_data.keys() and 'yaxis.range[0]' in relayout_data.keys():
        return True
    else:
        return False


def get_min_max_offset(df: pd.DataFrame, models: List[str]):
    min_offset = df[f'offset_{models[0]}'].min()
    max_offset = df[f'offset_{models[0]}'].max()

    if len(models) == 2:
        min_offset_m2 = df[f'offset_{models[1]}'].min()
        min_offset = min(min_offset, min_offset_m2)

        max_offset_m2 = df[f'offset_{models[1]}'].max()
        max_offset = max(max_offset, max_offset_m2)
    return min_offset, max_offset


def get_min_max_index(df: pd.DataFrame, models: List[str], y_start_idx: int, y_stop_idx: int):
    condition_y_min = (df[f'offset_{models[0]}'] >= y_start_idx)
    condition_y_max = (df[f'offset_{models[0]}'] <= y_stop_idx)
    df_temp = conditional_sliced_df(df, condition_y_min, condition_y_max)
    min_index = min(list(df_temp.index))
    max_index = max(list(df_temp.index))

    if len(models) == 2:
        condition_y_min_m2 = (df[f'offset_{models[1]}'] >= y_start_idx)
        condition_y_max_m2 = (df[f'offset_{models[1]}'] <= y_stop_idx)
        df_temp_m2 = conditional_sliced_df(df, condition_y_min_m2, condition_y_max_m2)
        min_index_m2 = min(list(df_temp_m2.index))
        min_index = min(min_index, min_index_m2)

        max_index_m2 = max(list(df_temp_m2.index))
        max_index = max(max_index, max_index_m2)
    return min_index, max_index


def get_min_max_cluster(df: pd.DataFrame, models: List[str], y_start_idx: int, y_stop_idx: int):
    condition_y_min = (df[f'offset_{models[0]}'] >= y_start_idx)
    condition_y_max = (df[f'offset_{models[0]}'] <= y_stop_idx)
    df_temp = conditional_sliced_df(df, condition_y_min, condition_y_max)
    min_cluster = df_temp[f'cluster_{models[0]}'].min()
    max_cluster = df_temp[f'cluster_{models[0]}'].max()

    if len(models) == 2:
        condition_y_min_m2 = (df[f'offset_{models[1]}'] >= y_start_idx)
        condition_y_max_m2 = (df[f'offset_{models[1]}'] <= y_stop_idx)
        df_temp_m2 = conditional_sliced_df(df, condition_y_min_m2, condition_y_max_m2)
        min_cluster_m2 = df_temp_m2[f'cluster_{models[1]}'].min()
        min_cluster = min(min_cluster, min_cluster_m2)

        max_cluster_m2 = df_temp_m2[f'cluster_{models[1]}'].max()
        max_cluster = max(max_cluster, max_cluster_m2)
    return min_cluster, max_cluster


def get_effective_xaxis_cluster(relayout_data: Dict):
    if relayout_data['xaxis.range[0]'] > 1 and relayout_data['xaxis.range[0]'] < 8:
        x_cluster = int(relayout_data['xaxis.range[0]']) + 1
    elif relayout_data['xaxis.range[0]'] <= 1:
        x_cluster = 1
    elif relayout_data['xaxis.range[0]'] >= 8:
        x_cluster = 8
    return x_cluster


def get_adjusted_dfs_based_on_legend_filtration(dfs: List[pd.DataFrame], models: List[str]):
    if models[0] != dfs[0]['model'].values[0]:
        dfs[0] = dfs[1]
    return dfs


def get_adjusted_xy_coordinate(relayout_data: Dict, df: pd.DataFrame):
    if detected_single_xaxis(relayout_data):
        x_start_idx = int(relayout_data['xaxis.range[0]']) if relayout_data['xaxis.range[0]'] >= 0 else 0
        x_stop_idx = int(relayout_data['xaxis.range[1]']) if relayout_data['xaxis.range[1]'] <= len(df) - 1 else len(df) - 1
        y_start_idx = 0
        y_stop_idx = 1

    elif detected_single_yaxis(relayout_data):
        x_start_idx = 0
        x_stop_idx = len(df) - 1
        y_start_idx = relayout_data['yaxis.range[0]'] if relayout_data['yaxis.range[0]'] >= 0 else 0
        y_stop_idx = relayout_data['yaxis.range[1]'] if relayout_data['yaxis.range[1]'] <= 1 else 1

    elif detected_complete_pair_xaxis_yaxis(relayout_data):
        x_start_idx = int(relayout_data['xaxis.range[0]']) if relayout_data['xaxis.range[0]'] >= 0 else 0
        x_stop_idx = int(relayout_data['xaxis.range[1]']) if relayout_data['xaxis.range[1]'] <= len(df) - 1 else len(df) - 1
        y_start_idx = relayout_data['yaxis.range[0]'] if relayout_data['yaxis.range[0]'] > 0 else 0
        y_stop_idx = relayout_data['yaxis.range[1]'] if relayout_data['yaxis.range[1]'] < 1 else 1
    return x_start_idx, x_stop_idx, y_start_idx, y_stop_idx


def get_max_value_on_slider(df: pd.DataFrame, component: str):
    if component == 'feat-dist':
        max_value = len(list(df.columns))
        if len(list(df.columns)) > 10:
            max_value = 10
        return max_value
    elif component == 'similaritiesCF':
        max_value = len(df)
        if len(df) > 10:
            max_value = 10
        return max_value


def dataframe_prep_on_model_count_by_yaxis_slice(df: pd.DataFrame, models: List[str], y_start_idx: int, y_stop_idx: int):
    '''
    For regression task
    '''
    offset_cols = [col for col in df.columns if 'offset_' in col]
    condition_m1_1 = (df[offset_cols[0]] >= y_start_idx)
    condition_m1_2 = (df[offset_cols[0]] <= y_stop_idx)

    if len(models) == 2:
        condition_m2_1 = (df[offset_cols[1]] >= y_start_idx)
        condition_m2_2 = (df[offset_cols[1]] <= y_stop_idx)
        df_final_m1 = conditional_sliced_df(df, condition_m1_1, condition_m1_2)
        df_final_m2 = conditional_sliced_df(df, condition_m2_1, condition_m2_2)

        final_filtered_idx = set(df_final_m1.index).union(set(df_final_m2.index))
        df_final = df[df['index'].isin(list(final_filtered_idx))]
    else:
        if offset_cols[0].replace('offset_', '') == models[0]:
            df_final = conditional_sliced_df(df, condition_m1_1, condition_m1_2)
        else:
            condition_m2_1 = (df[offset_cols[1]] >= y_start_idx)
            condition_m2_2 = (df[offset_cols[1]] <= y_stop_idx)
            df_final = conditional_sliced_df(df, condition_m2_1, condition_m2_2)
    return df_final


def new_dataframe_prep_based_on_effective_index(df: pd.DataFrame, df_ref: pd.DataFrame):
    '''
    For classification - loss cluster
    '''
    effective_index = list(df_ref['index'].unique())
    df_new = df[df['index'].isin(effective_index)]
    return df_new


def conditional_sliced_df(df: pd.DataFrame, condition1, condition2):
    df_sliced = df[condition1 & condition2]
    return df_sliced


def insert_index_col(df):
    df.insert(0, 'index', list(df.index))
    return df


def invalid_slicing_range(slicing_input: str):
    special_characters = list(string.punctuation.replace(':', '').replace('%', ''))
    letters = list(string.ascii_letters)
    try:
        if slicing_input.split(':') == ['']:  # when user does not enter any index range info
            invalid_1 = False
        else:
            invalid_1 = ':' not in slicing_input
    except AttributeError:  # 'NoneType' object has no attribute 'split' (NoneType occurs during first spin up of this app)
        invalid_1 = True

    invalid_2 = any(char in letters for char in slicing_input)
    invalid_3 = any(char in list(slicing_input) for char in special_characters)
    if any([invalid_1, invalid_2, invalid_3]) is True:
        return True
    else:
        return False


def invalid_slicing_range_sequence(range_list: List[str]):
    start = int(range_list[0].replace('%', ''))
    stop = int(range_list[1].replace('%', ''))
    if start > stop:
        return True
    else:
        return False


def invalid_min_max_limit(range_list: List[str], df: pd.DataFrame):
    start = int(range_list[0].replace('%', ''))
    stop = int(range_list[1].replace('%', ''))

    if start < 0 or stop > len(df):
        return True
    else:
        return False


def incomplete_range_entry(range_list: List[str]):
    '''
    Either only start_idx is provided or stop_idx is provided by user
    '''
    if (any(v == '' for v in range_list)) and (range_list.count('') != 2):
        return True
    else:
        return False


def _valid_fig_key_ls(relayout_data: Dict):
    valid_key_ls = []
    for rd in relayout_data:
        try:
            if rd is None:
                valid_key_ls.append(0)
            elif ('xaxis.range[0]' in rd.keys() or 'yaxis.range[0]' in rd.keys()) and 'xaxis.autorange' not in rd.keys():
                valid_key_ls.append(1)
        except KeyError:  # related to 'xaxis.range[0]'
            valid_key_ls.append(0)
    return valid_key_ls



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pandas as pd
import plotly.graph_objects as go
import plotly.express as px
import dash_table


def plot_probabilities_spread_pattern(df_specific_label: pd.DataFrame):
    '''
    Display scatter plot for probabilities comparison on correct data point vs miss-predicted data point
    for each class label

    Arguments:
        df_specific_label (:obj:`~pd.DataFrame`):
            dataframe of 1 specific label of 1 model type, output from int_miss_predictions

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying scatter plot outlining probabilities comparison on correct data point vs miss-predicted data point \
                for each class label
    '''
    label = list(df_specific_label.columns)[1]
    model_name = df_specific_label['model'].values[0]

    fig = px.scatter(df_specific_label,
                     x=list(df_specific_label.index),
                     y=df_specific_label[label],
                     color='pred_state',
                     category_orders={"pred_state": ["correct", "miss-predict"]},
                     color_discrete_sequence=px.colors.qualitative.D3)

    fig.update_layout(
        title=f'<b>Class {label}<br>[ {model_name} ]</b><br>',
        title_x=0.6,
        yaxis_title=f"Probability is_class_{label} ",
        yaxis_showgrid=False,
        xaxis_title="data_point index",
        xaxis_showgrid=False,
        legend_title="", 
        legend=dict(orientation="h", yanchor="bottom", y=1.03, xanchor="right", x=0.8), 
        width=250,
        height=600,
        margin=dict(t=170, b=0, l=12, r=12, pad=10))

    fig.update_yaxes(range=[0, 1])
    fig.update_xaxes(rangemode="tozero")
    fig.add_hline(y=0.5, line_dash="dot")

    # iterate through all traces, to ensure all label-class have consistent format
    for i in range(len(fig.data)):
        if fig.data[i]['legendgroup'] == 'correct':
            fig.data[i]['marker']['color'] = '#1f77b4'
            fig.data[i]['hovertemplate'] = "<b>Index %{x}</b><br>" + "<b>[ correct ]</b><br><br>" + \
                                            "probability: %{y:.4f}<br>" + "<extra></extra>"

        elif fig.data[i]['legendgroup'] == 'miss-predict':
            fig.data[i]['marker']['color'] = '#FF7F0E'
            fig.data[i]['hovertemplate'] = "<b>Index %{x}</b><br>" + "<b>[ miss-predict ]</b><br><br>" + \
                                            "probability: %{y:.4f}<br>" + "<extra></extra>"
    return fig


def plot_simple_probs_spread_overview(df_label_state: pd.DataFrame):
    '''
    Display data table listing simple stats on ss, %correct, % wrong, accuracy for each label class

    Arguments:
        df_label_state (:obj:`~pd.DataFrame`):
            dataframe containing info on simple stats, output from int_miss_predictions

    Returns:
        :obj:`~dash_table.DataTable`:
            table object outlining simple stats on ss, %correct, % wrong, accuracy for each label class
    '''
    fig = dash_table.DataTable(
        id='table', 
        columns=[{'id': c, 'name': c} for c in df_label_state.columns], 
        style_cell={'font-family': 'verdana', 
                    'font-size': '14px', 
                    'border': 'none', 
                    'minWidth': '100px'},
        style_header={'display': 'none'},
        style_table={'width': '550', 'margin': 'auto'},
        style_data={'lineHeight': '15px'},
        style_data_conditional=[{'if': {'column_id': 'index'}, 'textAlign': 'left'},
                                {'if': {'column_id': 'state_value'}, 'textAlign': 'right'}],
        data=df_label_state.to_dict('records'))
    return fig


def plot_prediction_offset_overview(df: pd.DataFrame):
    '''
    Display scatter plot for overview on prediction offset values

    Arguments:
        df (:obj:`~pd.DataFrame`):
            dataframe containing calculated offset values, output from int_miss_predictions

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying scatter plot outlining overview on prediction offset values by index
    '''
    pred_cols = [col for col in df.columns if 'yPred_' in col]
    offset_cols = [col for col in df.columns if 'offset_' in col]
    corrected_legend_names = [col.replace('yPred_', '') for col in pred_cols]
    df.insert(0, 'index', list(df.index))

    fig = px.scatter(df, x='index', y=offset_cols[0], custom_data=['index'], color_discrete_sequence=px.colors.qualitative.D3)
    fig.data[0].name = corrected_legend_names[0]
    fig.update_traces(showlegend=True, hovertemplate="Data Index : %{x}<br>Prediction Offset : %{y}")

    if len(pred_cols) > 1:  # Bimodal
        fig.add_trace(go.Scatter(
            x=df['index'], 
            y=df[offset_cols[1]], 
            name=corrected_legend_names[1], 
            mode='markers',
            marker=dict(color='#FF7F0E'),
            hovertemplate="Data Index : %{x}<br>Prediction Offset : %{y}"))

    # add reference baseline [mainly to have baseline included in legend]
    fig.add_trace(go.Scatter(
        x=[0, len(df)], 
        y=[0] * 2, 
        name="Baseline [Prediction - Actual]", 
        visible=True, 
        hoverinfo='skip',
        mode='lines',
        line=dict(color="green", dash="dot")))
    # referece baseline [mainly for the dotted line in graph, but no legend generated]
    fig.add_hline(y=0, line_dash="dot")

    fig.update_layout(
                title='<b>Prediction Offset Overview by Datapoint Index</b>', 
                xaxis_title='Datapoint Index', 
                yaxis_title='Offset from baseline', 
                title_x=0.5,
                legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1), 
                width=1000,
                height=550,
                margin=dict(t=110), 
                clickmode='event+select')

    return fig



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
import pandas as pd
import plotly.graph_objects as go
from rarity.utils.common_functions import is_regression


def plot_offset_clusters(df: pd.DataFrame, analysis_type: str):
    '''
    For use in regression task only.
    Function to plot figure displaying cluster groups by prediction offset values

    Arguments:
        df (:obj:`~pd.DataFrame`):
            dataframe containing cluster info, output from int_loss_clusters
        analysis_type (str):
            info to indicate if analysis is regression or classification, info inherited from data_loader

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying violin plot outlining cluster groups by offset values
    '''
    models = [col.replace('cluster_', '') for col in df.columns if 'cluster_' in col]
    fig = _plot_common_clusters(df, models, analysis_type)
    fig.update_layout(title_text='<b>Overview of Prediction Offset Clusters</b>', yaxis_title='Offset from baseline')
    return fig


def plot_logloss_clusters(dfs: List[pd.DataFrame], analysis_type: str):
    '''
    For use in classification task only.
    Function to plot figure displaying cluster groups by log-loss values

    Arguments:
        dfs (:obj:`List[~pd.DataFrame]`):
            list of dataframes containing cluster info, output from int_loss_clusters
        analysis_type (str):
            info to indicate if analysis is regression or classification, info inherited from data_loader

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying violin plot outlining cluster groups by log-loss values
    '''
    models = [df['model'].values[0] for df in dfs]
    fig = _plot_common_clusters(dfs, models, analysis_type)
    fig.update_layout(title_text='<b>Overview of Log-Loss Clusters on Miss-Prediction</b>', yaxis_title='Log-Loss')
    return fig


def plot_optimum_cluster_via_elbow_method(cluster_range: List[int], sum_squared_distance: List[float], models: List[str]):
    '''
    Figure to guide decision on the number of clusters that is reasonable to form with KMean method

    Arguments:
        cluster_range (:obj:`List[int]`):
            list of integers indicating the number of clusters
        sum_squared_distance (:obj:`List[float]`):
            list of sum of squared distance generated via kmean_inertia
        models (:obj:`List[str]`):
            list of models used to generate yPred

    Returns:
        :obj:`~plotly.graph_objects.Figure`:
            figure displaying line plot outlining the change in sum of squared distances along the cluster range
    '''
    fig = go.Figure()
    fig.add_trace(go.Scatter(x=cluster_range[0],
                            y=sum_squared_distance[0],
                            mode='lines+markers',
                            name=models[0],
                            line=dict(color='#1F77B4')))

    if len(cluster_range) == 2:
        fig.add_trace(go.Scatter(
                            x=cluster_range[1],
                            y=sum_squared_distance[1],
                            mode='lines+markers',
                            name=models[1],
                            line=dict(color='#FF7F0E')))

    fig.update_layout(title='<b>Overview of Optimum Cluster Count</b>',
                    xaxis_title='No. of Cluster',
                    yaxis_title='Sum of Squared Distance',
                    title_x=0.5,
                    showlegend=True,
                    legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1),
                    width=1000,
                    height=550)

    fig.update_xaxes(rangemode="tozero")
    return fig


def _plot_common_clusters(df: pd.DataFrame, models: List[str], analysis_type: str):
    '''
    Internal function to plot violin graph for different cluster groups
    '''
    fig = go.Figure()
    if is_regression(analysis_type):
        x_m1 = df[f'cluster_{models[0]}']
        y_m1 = df[f'offset_{models[0]}']
        customdata_m1 = list(df.index)
        if len(models) == 2:
            x_m2 = df[f'cluster_{models[1]}']
            y_m2 = df[f'offset_{models[1]}']
            customdata_m2 = customdata_m1

    else:  # binary classification
        x_m1 = df[0]['cluster']
        y_m1 = df[0]['lloss']
        customdata_m1 = list(df[0].index)
        if len(models) == 2:
            x_m2 = df[1]['cluster']
            y_m2 = df[1]['lloss']
            customdata_m2 = list(df[1].index)

    fig.add_trace(go.Violin(x=x_m1,
                            y=y_m1,
                            legendgroup=models[0], scalegroup=models[0], name=models[0],
                            line_color='#1f77b4',
                            customdata=customdata_m1,
                            hovertemplate='index=%{customdata}<br>cluster=%{x}<br>offset=%{y}<br>',
                            showlegend=True))

    if len(models) == 2:
        fig.add_trace(go.Violin(x=x_m2,
                                y=y_m2,
                                legendgroup=models[1], scalegroup=models[1], name=models[1],
                                line_color='#FF7F0E',
                                customdata=customdata_m2,
                                hovertemplate='index=%{customdata}<br>cluster=%{x}<br>offset=%{y}<br>',
                                showlegend=True))

    # update characteristics shared by all traces
    fig.update_traces(meanline_visible=True,
                    box_visible=True,
                    points='all',  # show all points
                    jitter=0.05,  # add some jitter on points for better visibility
                    scalemode='count')  # scale violin plot area with total count

    fig.update_layout(
        title_x=0.5,
        xaxis_title='Cluster',
        width=1000,
        height=550,
        legend=dict(orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1),
        violingap=0.2, violingroupgap=0.3, violinmode='overlay')
    return fig



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List
import math
import numpy as np
import pandas as pd
from scipy import stats
from sklearn.cluster import KMeans
from sklearn import metrics
from sklearn.metrics.pairwise import euclidean_distances


def create_clusters(loss_list: List[float], num_cluster: int):
    '''
    cluster creation for single model
    '''
    loss_array = np.array(loss_list).reshape(-1, 1)

    kmean = KMeans(n_clusters=num_cluster, random_state=42).fit(loss_array)

    # this lookup index is to ensure the same clustering results (cluster groups) even if Kmean randomized the initalization at diff centroid
    # reference: https://stackoverflow.com/questions/44888415/how-to-set-k-means-clustering-labels-from-highest-to-lowest-with-python
    idx = np.argsort(kmean.cluster_centers_.sum(axis=1))
    lookup = np.zeros_like(idx)
    lookup[idx] = np.arange(num_cluster)
    cluster_groups = lookup[kmean.labels_].tolist()

    # cluster_groups = kmean.labels_.tolist() ==> use this if don't mind different clustering groups at each button click
    cluster_groups = [cluster + 1 for cluster in cluster_groups]  # cluster to start from 1, instead of 0
    cluster_score = round(metrics.silhouette_score(loss_array, cluster_groups, metric='euclidean'), 4)
    return cluster_groups, cluster_score


def calculate_logloss(yTrue: float, yPred: float, log_func: math.log = math.log, eps: float = 1e-15):
    '''
    calculate single data point loss
    [sklearn] : Log loss is undefined for p=0 or p=1, so probabilities are clipped to max(eps, min(1 - eps, p))
    '''
    yPred = np.clip(yPred, eps, 1 - eps)
    loss_single_datapoint = -(-yTrue * log_func(yPred) + (1 - yTrue) * log_func(1 - yPred))
    return loss_single_datapoint


def get_optimum_num_clusters(loss_list: List[float], num_cluster: int):
    '''
    generate 2 output lists for cluster count and its corresponding sum of squared distance
    '''
    loss_array = np.array(loss_list).reshape(-1, 1)
    cluster_range = list(range(1, 10))

    sum_squared_distance = []
    for i in cluster_range:
        kmean = KMeans(n_clusters=i, random_state=42).fit(loss_array)
        sum_squared_distance.append(kmean.inertia_)
    return cluster_range, sum_squared_distance


def calculate_kl_div(seq1: Union[pd.DataFrame, List[float]], seq2: Union[pd.DataFrame, List[float]]):
    kl_divergence = stats.entropy(seq1, seq2)
    return kl_divergence


def get_optimum_bin_size(feature_pd_series: pd.Series):
    '''
    adopt Freedman-Diaconis rule
    '''
    q1 = feature_pd_series.quantile(0.25)
    q3 = feature_pd_series.quantile(0.75)
    inner_quantile = q3 - q1
    bin_width = (2 * inner_quantile) / (len(feature_pd_series) ** (1 / 3))
    bin_count = int(np.ceil((feature_pd_series.max() - feature_pd_series.min()) / bin_width))
    return bin_count


def compute_distances(X: pd.Series, Y: pd.Series, metric=euclidean_distances):
    return metric(X, Y)



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List
import pandas as pd

from rarity.data_loader import BaseLoader
from rarity.utils.common_functions import is_regression, is_classification


class CSVDataLoader(BaseLoader):
    '''
    Dataloader that compiles all input files in csv format.

    Args:
        xFeature_file (str):
            Path to csv file that contains all xfeatures used for model development/training.

            `Example of csv file storing xfeatures :`

            +-----------+-----------+
            | feature_0 | feature_1 |
            +-----------+-----------+
            | 21        | B         |
            +-----------+-----------+
            | 36        | A         |
            +-----------+-----------+


        yTrue_file (str):
            Path to csv file that contains all actual values (regression) / true labels (classification)

            `Example of csv file storing yTrue values for`

            ``Regression`` :

            +--------+
            | price  |
            +========+
            | 78634  |
            +--------+
            | 98273  |
            +--------+
            | 2780   |
            +--------+

            ``Binary Classification`` :

            +--------+
            | churn  |
            +========+
            | 1      |
            +--------+
            | 0      |
            +--------+
            | 1      |
            +--------+

            ``Multiclass Classification`` :

            +--------+
            | size   |
            +========+
            | big    |
            +--------+
            | medium |
            +--------+
            | small  |
            +--------+

        yPred_file_ls (:obj:`List[str]`):
            List consists of csv file paths that contain prediction values / probabilities generated by specific model type.
            One csv file for 1 model prediction outputs

            `Example of csv file stroing yPred values / labels for`

            ``Regression`` :

            +--------+
            | price  |
            +========+
            | 83683  |
            +--------+
            | 67293  |
            +--------+

            ``Binary Classification`` :

            +--------+--------+
            | 0      | 1      |
            +========+========+
            | 0.0675 | 0.9325 |
            +--------+--------+
            | 0.6237 | 0.3767 |
            +--------+--------+

            ``Multiclass Classification`` :

            +--------+--------+--------+
            | big    | medium | small  |
            +========+========+========+
            | 0.7772 | 0.1140 | 0.1088 |
            +--------+--------+--------+
            | 0.0014 | 0.8169 | 0.1817 |
            +--------+--------+--------+

        model_names_ls (:obj:`List[str]`):
            List contains model names representing the model used to generate yPred

        analysis_type (str):
            Analysis type defined by user. Corresponding feature components will be auto-populated based on the \
            specified analysis type. Supported analysis types : ``Regression``, ``Binary Classification``, ``Multiclass Classification``
    '''
    def __init__(
        self,
        xFeatures_file: str,
        yTrue_file: str,
        yPred_file_ls: List[str] = [],
        model_names_ls: List[str] = [],
        analysis_type: str = None
    ):
        super().__init__(xFeatures_file, yTrue_file, yPred_file_ls, model_names_ls, analysis_type)
        self.analysis_type = analysis_type.lower().replace(' ', '-')

        assert len(self.yPreds) == len(self.models), 'no. of yPred_files must be equal to \
                                                    no. of model_names in correct order'

        assert self.analysis_type in ['regression', 'binary-classification', 'multiclass-classification'], "Currently " \
            "supported analysis types: ['Regression', 'Binary Classification', 'Multiclass Classification']"

        if self.analysis_type == 'multiclass-classification':
            cols_ls = [list(pd.read_csv(yPred).columns) for yPred in self.yPreds]
            assert all(len(col) > 2 for col in cols_ls), "Data for yPred_file doesn't seem to be a multiclass prediction. " \
                "Please ensure there is prediction probabilities for each class in the yPred file."

        if self.analysis_type == 'binary-classification':
            assert pd.read_csv(self.yTrue).nunique().values[0] == 2, "Data for yTrue_file doesn't seem to be a binary-class " \
                "prediction. Please ensure yTrue_file consists of array of only 2 unique class"

    def get_features(self):
        return pd.read_csv(self.xFeatures)

    def get_yTrue(self):
        yTrue = pd.read_csv(self.yTrue)
        yTrue.rename(columns={yTrue.columns[0]: 'yTrue'}, inplace=True)
        return yTrue

    def get_yPreds(self):
        yPred_ls = [pd.read_csv(yPred) for yPred in self.yPreds]
        if is_regression(self.analysis_type):
            yPred_df = pd.concat(yPred_ls, axis=1)
            if len(self.models) > 1:
                yPred_df.rename(columns={yPred_df.columns[0]: f'yPred_{self.models[0]}', 
                                        yPred_df.columns[1]: f'yPred_{self.models[1]}'}, inplace=True)
            elif len(self.models) == 1:
                yPred_df.rename(columns={yPred_df.columns[0]: f'yPred_{self.models[0]}'}, inplace=True)
            return yPred_df
        elif is_classification(self.analysis_type):
            for i in range(len(self.models)):
                yPred_ls[i]['model'] = self.models[i]
                yPred_ls[i]['yPred-label'] = yPred_ls[i][list(yPred_ls[i].columns)[:-1]].idxmax(axis=1)
            return yPred_ls

    def get_model_list(self):
        return self.models

    def get_analysis_type(self):
        return self.analysis_type

    def get_all(self):
        if is_regression(self.analysis_type):
            df = pd.concat([self.get_features(), self.get_yTrue(), self.get_yPreds()], axis=1)
            return df
        elif is_classification(self.analysis_type):
            df_ls = []
            for i in range(len(self.models)):
                df = pd.concat([self.get_features(), self.get_yTrue(), self.get_yPreds()[i]], axis=1)
                df_ls.append(df)
    #         dfs = pd.concat(df_ls) # can remove this dfs if memory issue for big files
            return df_ls


class DataframeLoader(BaseLoader):
    '''
    Dataloader that compile all xFeatures, yTrue, yPreds in dataframe format.

    Args:
        df_xFeatures (:obj:`~pd.DataFrame`):
            Dataframe that contains all xfeatures used for model development/training.

        df_yTrue (:obj:`~pd.DataFrame`):
            Dataframe that contains all true values / labels.

        df_yPred_ls (:obj:`List[~pd.DataFrame]`):
            Dataframe that contains all predicted values (regresession) / probabilities (classification).

        model_names_ls (:obj:`List[str]`):
            List contains model names representing the model used to generate yPred.

        analysis_type (str):
            Analysis type defined by user. Corresponding feature components will be auto-populated based on the
            specified analysis type. Supported analysis types : ``Regression``, ``Binary Classification``, ``Multiclass Classification``

    '''
    def __init__(
        self,
        df_xFeatures: pd.DataFrame,
        df_yTrue: pd.DataFrame,
        df_yPred_ls: List[pd.DataFrame] = [],
        model_names_ls: List[str] = [],
        analysis_type: str = None
    ):
        super().__init__(df_xFeatures, df_yTrue, df_yPred_ls, model_names_ls, analysis_type)
        self.yTrue = df_yTrue.copy()
        self.yPreds = [df.copy() for df in df_yPred_ls]
        self.analysis_type = analysis_type.lower().replace(' ', '-')
        assert len(self.yPreds) == len(self.models), 'no. of yPred_files must be equal to ' \
                'no. of model_names in correct order'

        assert self.analysis_type in ['regression', 'binary-classification', 'multiclass-classification'], "Currently " \
            "supported analysis types: ['Regression', 'Binary Classification', 'Multiclass Classification']"

        if self.analysis_type == 'multiclass-classification':
            cols_ls = [list(yPred.columns) for yPred in self.yPreds]
            assert all(len(col) > 2 for col in cols_ls), "Data for df_yPred_list doesn't seem to be a multiclass prediction. " \
                "Please ensure there is prediction probabilities for each class in the df of df_yPred_list."

        if self.analysis_type == 'binary-classification':
            assert self.yTrue.nunique().values[0] == 2, "Data for df_yTrue doesn't seem to be a binary-class prediction. " \
                "Please ensure df_yTrue consists of data with only 2 unique class"

    def get_features(self):
        return self.xFeatures

    def get_yTrue(self):
        self.yTrue.rename(columns={self.yTrue.columns[0]: 'yTrue'}, inplace=True)
        return self.yTrue

    def get_yPreds(self):
        if is_regression(self.analysis_type):
            yPred_df = pd.concat(self.yPreds, axis=1)
            if len(self.models) > 1:
                yPred_df.rename(columns={yPred_df.columns[0]: f'yPred_{self.models[0]}', 
                                        yPred_df.columns[1]: f'yPred_{self.models[1]}'}, inplace=True)
            elif len(self.models) == 1:
                yPred_df.rename(columns={yPred_df.columns[0]: f'yPred_{self.models[0]}'}, inplace=True)
            return yPred_df
        elif is_classification(self.analysis_type):
            for i in range(len(self.models)):
                self.yPreds[i]['model'] = self.models[i]
                self.yPreds[i]['yPred-label'] = self.yPreds[i][list(self.yPreds[i].columns)[:2]].idxmax(axis=1)
            return self.yPreds

    def get_model_list(self):
        return self.models

    def get_analysis_type(self):
        return self.analysis_type

    def get_all(self):
        if is_regression(self.analysis_type):
            df = pd.concat([self.get_features(), self.get_yTrue(), self.get_yPreds()], axis=1)
            return df
        elif is_classification(self.analysis_type):
            df_ls = []
            for i in range(len(self.models)):
                df = pd.concat([self.xFeatures, self.get_yTrue(), self.get_yPreds()[i]], axis=1)
                df_ls.append(df)
    #         dfs = pd.concat(df_ls) # can remove this dfs if memory issue for big files
            return df_ls



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List, Dict
import math
import pandas as pd

import dash
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
import dash_table
import dash_core_components as dcc
import dash_html_components as html
import dash_bootstrap_components as dbc

from rarity.app import app
from rarity.data_loader import CSVDataLoader, DataframeLoader
from rarity.interpreters.structured_data import IntLossClusterer
from rarity.visualizers import loss_clusters as viz_clusters
from rarity.visualizers import shared_viz_component as viz_shared
from rarity.utils import style_configs
from rarity.utils.common_functions import (is_active_trace, is_reset, is_regression, is_classification,
                                            detected_legend_filtration, detected_single_xaxis, detected_single_yaxis,
                                            detected_bimodal, get_min_max_offset, get_min_max_cluster, get_effective_xaxis_cluster,
                                            get_adjusted_dfs_based_on_legend_filtration, conditional_sliced_df, insert_index_col,
                                            dataframe_prep_on_model_count_by_yaxis_slice, new_dataframe_prep_based_on_effective_index)


def fig_plot_offset_clusters_reg(data_loader: Union[CSVDataLoader, DataframeLoader], num_cluster: int):
    '''
    For use in regression task only.
    Function to output collated info packs used to display final graph objects by cluster groups along with calculated silhouette scores

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module
        num_cluster (int):
            Number of cluster to form

    Returns:

             Compact outputs consist of the followings

            - df (:obj:`~pd.DataFrame`): dataframes for overview visualization need with offset values included
            - fig_obj_cluster (:obj:`~plotly.graph_objects.Figure`): figure displaying violin plot outlining cluster groups by offset values
            - ls_cluster_score (:obj:`List[str]`): list of silhouette scores, indication of clustering quality
            - fig_obj_elbow (:obj:`~plotly.graph_objects.Figure`): figure displaying line plot outlining the change in sum of squared distances \
                along the cluster range
    '''
    df, ls_cluster_score, ls_cluster_range, ls_ssd = IntLossClusterer(data_loader).xform(num_cluster, None, 'All')
    models = data_loader.get_model_list()
    analysis_type = data_loader.get_analysis_type()

    fig_obj_cluster = viz_clusters.plot_offset_clusters(df, analysis_type)
    fig_obj_elbow = viz_clusters.plot_optimum_cluster_via_elbow_method(ls_cluster_range, ls_ssd, models)
    return df, fig_obj_cluster, ls_cluster_score, fig_obj_elbow


def fig_plot_logloss_clusters_cls(data_loader: Union[CSVDataLoader, DataframeLoader],
                                    num_cluster: int,
                                    log_func: math.log = math.log,
                                    specific_dataset: str = 'All'):
    '''
    For use in classification task only.
    Function to output collated info packs used to display final graph objects by cluster groups along with calculated silhouette scores

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module
        num_cluster (int):
            Number of cluster to form
        log_funct (:obj:`math.log`):
            Mathematics logarithm function used to calculate log-loss between yTrue and yPred
        specific_dataset (str):
            Default to 'All' indicating to include all miss-predict labels. Other options flexibly expand depending on class labels

    Returns:

             Compact outputs consist of the followings

            - ls_dfs_viz (:obj:`List[~pd.DataFrame]`): dataframes for overview visualization need with offset values included
            - fig_obj_cluster (:obj:`~plotly.graph_objects.Figure`): figure displaying violin plot outlining cluster groups by offset values
            - ls_cluster_score (:obj:`List[str]`): list of silhouette scores, indication of clustering quality
            - fig_obj_elbow (:obj:`~plotly.graph_objects.Figure`): figure displaying line plot outlining the change in sum of squared distances \
                along the cluster range
            - ls_class_labels (:obj:`List[str]`): list of all class labels
            - ls_class_labels_misspred (:obj:`List[str]`): list of class labels with minimum of 1 miss-prediction
            - df_features (:obj:`~pandas.DataFrame`): dataframe storing all features used in dataset
    '''
    compact_outputs = IntLossClusterer(data_loader).xform(num_cluster, log_func, specific_dataset)
    ls_dfs_viz, ls_class_labels, ls_class_labels_misspred = compact_outputs[0], compact_outputs[1], compact_outputs[2]
    ls_cluster_score, ls_cluster_range, ls_ssd = compact_outputs[3], compact_outputs[4], compact_outputs[5]
    df_features = data_loader.get_features()
    analysis_type = data_loader.get_analysis_type()
    models = data_loader.get_model_list()

    fig_obj_cluster = viz_clusters.plot_logloss_clusters(ls_dfs_viz, analysis_type)
    fig_obj_elbow = viz_clusters.plot_optimum_cluster_via_elbow_method(ls_cluster_range, ls_ssd, models)
    return ls_dfs_viz, fig_obj_cluster, ls_cluster_score, fig_obj_elbow, ls_class_labels, ls_class_labels_misspred, df_features


def table_with_relayout_datapoints(data: dash_table.DataTable, customized_cols: List[str], header: Dict, exp_format: str):
    '''
    Create table outlining dataframe content

    Arguments:
        data (:obj:`~dash_table.DataTable`):
            dictionary like format storing dataframe info under 'record' key
        customized_cols (:obj:`List[str]`):
            list of customized column names
        header (:obj:`Dict`):
            dictionary format storing the style info for table header
        exp_format (str):
            text info indicating the export format

    Returns:
        :obj:`~dash_table.DataTable`:
            table object outlining the dataframe content with specific styles
    '''
    tab_obj = viz_shared.reponsive_table_to_filtered_datapoints(data, customized_cols, header, exp_format)
    return tab_obj


def convert_cluster_relayout_data_to_df_reg(relayout_data: Dict, df: pd.DataFrame, models: List[str]):
    '''
    For use in regression task only.
    Convert raw data format from relayout selection range by user into the correct df fit for viz purpose

    Arguments:
        relayout_data (:obj:`Dict`):
            dictionary like data containing selection range indices returned from plotly graph
        df (:obj:`~pandas.DataFrame`):
            dataframe tap-out from interpreters pipeline
        models (:obj:`List[str]`):
            model names defined by user during spin-up of Tenjin app

    Returns:
        :obj:`~pandas.DataFrame`:
            dataframe fit for the responsive table-graph filtering
    '''
    if detected_single_xaxis(relayout_data):
        x_cluster = get_effective_xaxis_cluster(relayout_data)
        df_filtered_x = df[df[f'cluster_{models[0]}'] == x_cluster]

        if detected_bimodal(models):
            df_filtered_x_m2 = df[df[f'cluster_{models[1]}'] == x_cluster]
            df_filtered_x = pd.concat([df_filtered_x, df_filtered_x_m2]).drop_duplicates()

        y_start_idx, y_stop_idx = get_min_max_offset(df_filtered_x, models)
        df_final = dataframe_prep_on_model_count_by_yaxis_slice(df_filtered_x, models, y_start_idx, y_stop_idx)

    elif detected_single_yaxis(relayout_data):
        y_start_idx = relayout_data['yaxis.range[0]']
        y_stop_idx = relayout_data['yaxis.range[1]']
        df_filtered_y = dataframe_prep_on_model_count_by_yaxis_slice(df, models, y_start_idx, y_stop_idx)

        x_start_idx, x_stop_idx = get_min_max_cluster(df_filtered_y, models, y_start_idx, y_stop_idx)
        x_start_idx = x_start_idx if x_start_idx >= 1 else 1
        x_stop_idx = x_stop_idx if x_stop_idx <= 8 else 8

        condition_min_cluster = df_filtered_y[f'cluster_{models[0]}'] >= x_start_idx
        condition_max_cluster = df_filtered_y[f'cluster_{models[0]}'] <= x_stop_idx
        df_final = conditional_sliced_df(df_filtered_y, condition_min_cluster, condition_max_cluster)

        if detected_bimodal(models):
            condition_min_cluster_m2 = df_filtered_y[f'cluster_{models[1]}'] >= x_start_idx
            condition_max_cluster_m2 = df_filtered_y[f'cluster_{models[1]}'] <= x_stop_idx
            df_final_m2 = conditional_sliced_df(df_filtered_y, condition_min_cluster_m2, condition_max_cluster_m2)
            df_final = pd.concat([df_final, df_final_m2]).drop_duplicates()

    else:  # a complete range is provided by user (with proper x-y coordinates)
        x_cluster = get_effective_xaxis_cluster(relayout_data)
        y_start_idx = relayout_data['yaxis.range[0]']
        y_stop_idx = relayout_data['yaxis.range[1]']
        df_filtered = df[df[f'cluster_{models[0]}'] == x_cluster]

        if detected_bimodal(models):
            df_filtered_m2 = df[df[f'cluster_{models[1]}'] == x_cluster]
            df_filtered = pd.concat([df_filtered, df_filtered_m2]).drop_duplicates()

        df_final = dataframe_prep_on_model_count_by_yaxis_slice(df_filtered, models, y_start_idx, y_stop_idx)
    return df_final


def convert_cluster_relayout_data_to_df_cls(relayout_data: Dict, dfs_viz: List[pd.DataFrame], df_features: pd.DataFrame, models: List[str]):
    '''
    For use in classification task only.
    Convert raw data format from relayout selection range by user into the correct df fit for viz purpose

    Arguments:
        relayout_data (:obj:`Dict`):
            dictionary like data containing selection range indices returned from plotly graph
        dfs_viz (:obj:`List[~pd.DataFrame]`):
            list of dataframes for overview visualization need with offset values included
        df_features (:obj:`~pandas.DataFrame`):
            dataframe storing all features used in dataset
        models (:obj:`List[str]`):
            model names defined by user during spin-up of Tenjin app

    Returns:

             Compact outputs consist of the followings

            - df_final_features (:obj:`~pd.DataFrame`): dataframe storing all features based on slicing info from relayout_data
            - df_final_probs (:obj:`~pd.DataFrame`): dataframe storing probability values by class label corresponding to \
                the slicing relayout_data
    '''
    if detected_single_xaxis(relayout_data):
        x_cluster = get_effective_xaxis_cluster(relayout_data)
        df_final_probs = dfs_viz[0][dfs_viz[0]['cluster'] == x_cluster]

        if detected_bimodal(models):
            df_final_probs_m2 = dfs_viz[1][dfs_viz[1]['cluster'] == x_cluster]
            df_final_probs = pd.concat([df_final_probs, df_final_probs_m2]).drop_duplicates()
            df_final_probs = df_final_probs.sort_values('index')  # so that index of different models will appear together row-row

        df_final_features = new_dataframe_prep_based_on_effective_index(df_features, df_final_probs)

    elif detected_single_yaxis(relayout_data):
        y_start_idx = relayout_data['yaxis.range[0]']
        y_stop_idx = relayout_data['yaxis.range[1]']

        condition_min_loss = dfs_viz[0]['lloss'] >= y_start_idx
        condition_max_loss = dfs_viz[0]['lloss'] <= y_stop_idx
        df_final_probs = conditional_sliced_df(dfs_viz[0], condition_min_loss, condition_max_loss)

        if detected_bimodal(models):
            condition_min_loss_m2 = dfs_viz[1]['lloss'] >= y_start_idx
            condition_max_loss_m2 = dfs_viz[1]['lloss'] <= y_stop_idx
            df_final_probs_m2 = conditional_sliced_df(dfs_viz[1], condition_min_loss_m2, condition_max_loss_m2)

            df_final_probs = pd.concat([df_final_probs, df_final_probs_m2]).drop_duplicates()
            df_final_probs = df_final_probs.sort_values('index')  # so that index of different models will appear together row-row

        df_final_features = new_dataframe_prep_based_on_effective_index(df_features, df_final_probs)

    else:  
        '''
        detected_single_xaxis or a complete range is provided by user (with proper x-y coordinates) 
        will have same results due to the setup of dfs_viz for cls (loss values are tight to cluster group)
        '''
        x_cluster = get_effective_xaxis_cluster(relayout_data)
        df_filtered_x = dfs_viz[0][dfs_viz[0]['cluster'] == x_cluster]

        y_start_idx = relayout_data['yaxis.range[0]']
        y_stop_idx = relayout_data['yaxis.range[1]']
        condition_min_loss = df_filtered_x['lloss'] >= y_start_idx
        condition_max_loss = df_filtered_x['lloss'] <= y_stop_idx
        df_final_probs = conditional_sliced_df(df_filtered_x, condition_min_loss, condition_max_loss)

        if detected_bimodal(models):
            df_filtered_x_m2 = dfs_viz[1][dfs_viz[1]['cluster'] == x_cluster]
            condition_min_loss_m2 = df_filtered_x_m2['lloss'] >= y_start_idx
            condition_max_loss_m2 = df_filtered_x_m2['lloss'] <= y_stop_idx
            df_final_probs_m2 = conditional_sliced_df(df_filtered_x_m2, condition_min_loss_m2, condition_max_loss_m2)

            df_final_probs = pd.concat([df_final_probs, df_final_probs_m2]).drop_duplicates()
            df_final_probs = df_final_probs.sort_values('index')  # so that index of different models will appear together row-row

        df_final_features = new_dataframe_prep_based_on_effective_index(df_features, df_final_probs)
    return df_final_features, df_final_probs


def _display_score(ls_cluster_score: List[float], models: List[str]):
    '''
    Internal function to tap-out text field for silhouette score
    '''
    score_text = f'Silhouette score: {ls_cluster_score[0]}'
    if detected_bimodal(models):
        score_text = f'Silhouette score: {ls_cluster_score[0]} [ {models[0]} ] ' \
                    f'{ls_cluster_score[1]} [ {models[1]} ]'
    return score_text


class LossClusters:
    '''
    Main integration for feature component on Loss Clusters.

    Arguments:
        data_loader (:class:`~rarity.data_loader.CSVDataLoader` or :class:`~rarity.data_loader.DataframeLoader`):
            Class object from data_loader module

    Important Attributes:

        analysis_type (str):
            Analysis type defined by user during initial inputs preparation via data_loader stage.
        model_names (:obj:`List[str]`):
            model names defined by user during initial inputs preparation via data_loader stage.
        is_bimodal (bool):
            to indicate if analysis involves 2 models
        num_clusters (int):
            Number of cluster to form
        log_funct (:obj:`math.log`):
            Mathematics logarithm function used to calculate log-loss between yTrue and yPred
        specific_dataset (str):
            Default to 'All' indicating to include all miss-predict labels. Other options flexibly expand depending on class labels

    Returns:
        :obj:`~dash_core_components.Container`:
            styled dash components displaying graph and/or table objects
    '''
    def __init__(self, data_loader: Union[CSVDataLoader, DataframeLoader]):
        self.data_loader = data_loader
        self.analysis_type = data_loader.get_analysis_type()
        self.model_names = data_loader.get_model_list()
        self.is_bimodal = True if len(self.model_names) > 1 else False
        self.num_cluster = 4

        # instantiate at this stage due to shared use with callbacks
        if is_regression(self.analysis_type):
            self.compact_outputs_reg = fig_plot_offset_clusters_reg(self.data_loader, self.num_cluster)
            self.df, self.offset_clusters_reg = self.compact_outputs_reg[0], self.compact_outputs_reg[1]
            self.ls_cluster_score, self.optimum_elbow_reg = self.compact_outputs_reg[2], self.compact_outputs_reg[3]
            self.cols_table_reg = [col.replace('_', ' ') for col in self.df.columns]
            self.score_text = _display_score(self.ls_cluster_score, self.model_names)

        elif is_classification(self.analysis_type):
            self.log_func = math.log
            self.specific_dataset = 'All'

            self.compact_outputs_cls = fig_plot_logloss_clusters_cls(self.data_loader, self.num_cluster, self.log_func, self.specific_dataset)
            self.ls_dfs_viz, self.lloss_clusters_cls, = self.compact_outputs_cls[0], self.compact_outputs_cls[1]
            self.ls_cluster_score, self.optimum_elbow_cls = self.compact_outputs_cls[2], self.compact_outputs_cls[3]
            self.ls_class_labels, self.ls_class_labels_misspred = self.compact_outputs_cls[4], self.compact_outputs_cls[5]
            self.score_text = _display_score(self.ls_cluster_score, self.model_names)

    def show(self):
        '''
        Method to tapout styled html for loss clusters
        '''
        if is_regression(self.analysis_type):
            lloss_clusters = dbc.Container([
                                    dbc.Row(html.Div(
                                        html.H5('Optimum Cluster via Elbow Method', className='h5__cluster-section-title'))),
                                    dbc.Row(
                                        dcc.Graph(id='fig-optimum-cluster-reg',
                                                figure=self.optimum_elbow_reg,),
                                        justify='center', className='border__optimum-cluster'),
                                    dbc.Row(html.H5('Log-Loss Clustering via KMean', className='h5__cluster-section-title')),
                                    dbc.Row([
                                            dbc.Col([
                                                dbc.Row(html.Div(html.H6('Select No. of Cluster'), className='h6__cluster-instruction')),
                                                dbc.Row(dbc.Select(id='select-num-cluster-reg',
                                                            options=style_configs.OPTIONS_NO_OF_CLUSTERS,
                                                            value='4'), className='params__select-cluster')
                                            ], width=6),
                                            dbc.Col(width=4),
                                            dbc.Col(
                                                dbc.Row(dcc.Loading(id='loading-output-loss-cluster-reg',
                                                                    type='circle', color='#a80202'),
                                                        justify='right', className='loading__loss-cluster'), width=1),
                                            dbc.Col(
                                                dbc.Row(dbc.Button("Update",
                                                                    id='button-num-cluster-update-reg',
                                                                    n_clicks=0,
                                                                    color="info", 
                                                                    className='button__update-dataset'),
                                                        justify='right'))], className='border__select-dataset'),
                                    dbc.Row(dbc.Col(dbc.Row(
                                        html.Div(self.score_text,
                                                id='text-score-cluster-reg',
                                                className='text__score-cluster-reg'), justify='right'))),
                                    dbc.Row(
                                        dcc.Graph(id='fig-loss-cluster-reg',
                                                    figure=self.offset_clusters_reg),
                                        justify='center', className='border__common-cluster-plot-reg'),

                                    html.Div(html.H6(style_configs.INSTRUCTION_TEXT_SHARED), className='h6__dash-table-instruction-cluster-reg'),
                                    html.Div(id='alert-to-reset-cluster-reg'),
                                    html.Div(id='table-feat-prob-cluster-reg', className='div__table-proba-misspred'),
                                    html.Br()], fluid=True)
            return lloss_clusters

        elif is_classification(self.analysis_type):
            options_misspred_dataset = [{'label': 'All', 'value': 'All'}] + \
                                        [{'label': f'class {label}', 'value': f'class {label}'} for label in self.ls_class_labels_misspred]

            lloss_clusters = dbc.Container([
                                    dbc.Row(html.H5('Loss Cluster Analysis for ALL Miss Predictions',
                                        id='title-after-misspred-dataset-selection-cls',
                                        className='h5__cluster-section-title')),
                                    dbc.Row([
                                        dbc.Col([
                                            dbc.Row(html.Div(html.H6('Select Miss-Predict Dataset'), className='h6__cluster-instruction')),
                                            dbc.Row(dbc.Select(id='select-misspred-dataset-cls',
                                                        options=options_misspred_dataset,
                                                        value='All'), className='params__select-cluster')
                                        ], width=4),
                                        dbc.Col(width=6),
                                        dbc.Col(
                                                dbc.Row(dcc.Loading(id='loading-output-misspred-dataset-cls',
                                                                    type='circle', color='#a80202'),
                                                        justify='right', className='loading__loss-cluster'), width=1),
                                        dbc.Col(
                                            dbc.Row(dbc.Button("Update",
                                                                id='button-misspred-dataset-update-cls',
                                                                n_clicks=0,
                                                                color="info",
                                                                className='button__update-dataset'),
                                                    justify='right'))], className='border__select-dataset'),
                                    html.Div(id='alert-clustering-error-cls'),
                                    dbc.Row(dcc.Graph(id='fig-cls-optimum-cluster', figure=self.optimum_elbow_cls,),
                                        justify='center', className='border__optimum-cluster'),
                                    dbc.Row(html.H5('Log-Loss Clustering via KMean on ALL Miss Predictions',
                                                    id='title-after-losscluster-params-selection-cls',
                                                    className='h5__cluster-section-title-kmean')),
                                    dbc.Row([
                                        dbc.Col([
                                            dbc.Row(html.Div(html.H6('Select No. of Cluster'), className='h6__cluster-instruction')),
                                            dbc.Row(dbc.Select(id='select-num-cluster-cls',
                                                        options=style_configs.OPTIONS_NO_OF_CLUSTERS,
                                                        value='4'), className='params__select-cluster')
                                        ], width=4),
                                        dbc.Col([
                                            dbc.Row(html.Div(html.H6('Select Logarithm Method'), className='h6__cluster-instruction')),
                                            dbc.Row(dbc.Select(id='select-log-method-cls',
                                                        options=[{'label': 'LOG', 'value': 'log'},
                                                                {'label': 'LOG1P', 'value': 'log1p'},
                                                                {'label': 'LOG2', 'value': 'log2'},
                                                                {'label': 'LOG10', 'value': 'log10'}],
                                                        value='log'), className='params__select-cluster')
                                        ], width=4),
                                        dbc.Col(width=2),
                                        dbc.Col(
                                                dbc.Row(dcc.Loading(id='loading-output-loss-cluster-cls',
                                                                    type='circle', color='#a80202'),
                                                        justify='right', className='loading__loss-cluster'), width=1),
                                        dbc.Col(
                                            dbc.Row(dbc.Button("Update",
                                                                id='button-logloss-update-cls',
                                                                n_clicks=0,
                                                                color="info",
                                                                className='button__update-dataset'),
                                                    justify='right'))], className='border__select-dataset'),
                                    dbc.Row(dbc.Col(dbc.Row(
                                        html.Div(self.score_text,
                                                id='text-score-cluster-cls',
                                                className='text__score-cluster-cls'), justify='right'))),
                                    dbc.Row(
                                        dcc.Graph(id='fig-loss-cluster-cls',
                                                    figure=self.lloss_clusters_cls,),
                                        justify='center', className='border__common-cluster-plot-cls'),

                                    # data-table, appeared only after data range selection on fig-loss-cluster-cls by user
                                    html.Div(html.H6(style_configs.INSTRUCTION_TEXT_SHARED), className='h6__dash-table-instruction-cls'),
                                    html.Div(id='alert-to-reset-loss-cluster-cls'),
                                    html.Div(id='table-title-features-loss-cluster'),
                                    html.Div(id='show-feat-table-loss-cluster', className='div__table-proba-misspred'),
                                    html.Br(),
                                    html.Div(id='table-title-probs-loss-cluster'),
                                    html.Div(id='show-prob-table-loss-cluster', className='div__table-proba-misspred'),
                                    html.Br()], fluid=True)
            return lloss_clusters

    def callbacks(self):
        @app.callback(
            Output('loading-output-loss-cluster-reg', 'children'),
            Output('text-score-cluster-reg', 'children'),
            Output('fig-loss-cluster-reg', 'figure'),
            Input('button-num-cluster-update-reg', 'n_clicks'),
            State('select-num-cluster-reg', 'value'))
        def update_fig_based_on_selected_num_cluster(click_count, selected_no_cluster):
            '''
            Callbacks functionalities specific to param - select no. of clusters [ regression ]
            '''
            if click_count > 0:
                _, fig_obj_cluster_reg, ls_cluster_score_reg, _ = fig_plot_offset_clusters_reg(self.data_loader, int(selected_no_cluster))
                score_text_reg = _display_score(ls_cluster_score_reg, self.model_names)
                return '', score_text_reg, fig_obj_cluster_reg
            else:
                raise PreventUpdate

        @app.callback(
            Output('alert-to-reset-cluster-reg', 'children'),
            Output('table-feat-prob-cluster-reg', 'children'),
            Input('fig-loss-cluster-reg', 'relayoutData'),
            Input('fig-loss-cluster-reg', 'restyleData'),
            State('select-num-cluster-reg', 'value'))
        def display_table_based_on_selected_range_reg(relayout_data, restyle_data, selected_no_cluster):
            '''
            Callbacks functionalities specific to reponse from fig-obj to data-table [ regression ]
            '''
            if relayout_data is not None:
                df_usr_select_cluster, _, _, _ = fig_plot_offset_clusters_reg(self.data_loader, int(selected_no_cluster))
                try:
                    # to limit table cell having values with long decimals for better viz purpose
                    df_usr_select_cluster = df_usr_select_cluster.round(2)
                except TypeError:
                    df_usr_select_cluster

                if is_reset(relayout_data):
                    alert_obj_reg = None
                    table_obj_reg = None

                elif is_active_trace(relayout_data):
                    models = self.model_names
                    if restyle_data is not None:  # [{'visible': ['legendonly']}, [1]]
                        if detected_legend_filtration(restyle_data):
                            model_to_exclude_from_view = self.model_names[restyle_data[1][0]]
                            models = [model for model in self.model_names if model != model_to_exclude_from_view]

                    default_header = style_configs.default_header_style()
                    alert_obj_reg = style_configs.activate_alert()

                    df_final = convert_cluster_relayout_data_to_df_reg(relayout_data, df_usr_select_cluster, models)
                    df_final.columns = self.cols_table_reg  # to have customized column names displayed on table

                    data_relayout_reg = df_final.to_dict('records')
                    table_obj_reg = table_with_relayout_datapoints(data_relayout_reg, self.cols_table_reg, default_header, 'csv')
                return alert_obj_reg, table_obj_reg
            else:
                raise PreventUpdate

        @app.callback(
            Output('loading-output-misspred-dataset-cls', 'children'),
            Output('loading-output-loss-cluster-cls', 'children'),
            Output('alert-clustering-error-cls', 'children'),
            Output('title-after-misspred-dataset-selection-cls', 'children'),
            Output('fig-cls-optimum-cluster', 'figure'),
            Output('title-after-losscluster-params-selection-cls', 'children'),
            Output('text-score-cluster-cls', 'children'),
            Output('fig-loss-cluster-cls', 'figure'),
            Input('button-misspred-dataset-update-cls', 'n_clicks'),
            Input('button-logloss-update-cls', 'n_clicks'),
            State('select-misspred-dataset-cls', 'value'),
            State('select-num-cluster-cls', 'value'),
            State('select-log-method-cls', 'value'))
        def update_loss_cluster_tab_based_on_selected_misspred_dataset(click_count_dataset,
                                                                        click_count_params,
                                                                        selected_dataset,
                                                                        selected_cluster,
                                                                        selected_method):
            '''
            Callbacks functionalities specific to all params selection [ classification ]
            '''
            ctx = dash.callback_context
            triggered_button = ctx.triggered[0]['prop_id'].split('.')[0]
            triggered_button_value = ctx.triggered[0]['value']

            current_dataset_name = ctx.states['select-misspred-dataset-cls.value']
            specific_dataset = current_dataset_name.replace('class ', '') if 'class' in current_dataset_name else current_dataset_name
            cluster_err_alert = style_configs.no_error_alert()

            # for click action on dataset selection
            if (triggered_button == 'button-misspred-dataset-update-cls') and (triggered_button_value > 0):
                title_aft_misspred_dataset = f'Loss Cluster Analysis for {selected_dataset.capitalize()} Miss Predictions'
                title_aft_params = f'Log-Loss Clustering via KMean on {selected_dataset.capitalize()} Miss Predictions'

                # pre-requisite to check if dataset is valid with sufficient data-points for auto-clustering
                ls_dfs_prob_misspred = IntLossClusterer(self.data_loader).extract_misspredictions()
                if specific_dataset != 'All' and any(len(df[df['yPred-label'] == specific_dataset]) < 8 for df in ls_dfs_prob_misspred):
                    cluster_err_alert = style_configs.activate_cluster_error_alert(specific_dataset)
                    return dash.no_update, dash.no_update, cluster_err_alert, dash.no_update, \
                            dash.no_update, dash.no_update, dash.no_update, dash.no_update

                outputs_callback_dataset = fig_plot_logloss_clusters_cls(self.data_loader,
                                                                        num_cluster=int(selected_cluster),
                                                                        log_func=style_configs.LOG_METHOD_DICT[selected_method],
                                                                        specific_dataset=specific_dataset)

                fig_obj_cluster_cls, ls_cluster_score_cls = outputs_callback_dataset[1], outputs_callback_dataset[2]
                fig_obj_elbow_cls = outputs_callback_dataset[3]
                text_score_cls = _display_score(ls_cluster_score_cls, self.model_names)
                return '', dash.no_update, cluster_err_alert, title_aft_misspred_dataset, fig_obj_elbow_cls, \
                        title_aft_params, text_score_cls, fig_obj_cluster_cls

            # for click action on num_cluster and log_method selection
            elif (triggered_button == 'button-logloss-update-cls') and (triggered_button_value > 0):
                outputs_callback_params = fig_plot_logloss_clusters_cls(self.data_loader,
                                                                        num_cluster=int(selected_cluster),
                                                                        log_func=style_configs.LOG_METHOD_DICT[selected_method],
                                                                        specific_dataset=specific_dataset)

                fig_obj_cluster_cls_params, ls_cluster_score_cls_params = outputs_callback_params[1], outputs_callback_params[2]
                text_score_cls_params = _display_score(ls_cluster_score_cls_params, self.model_names)
                return dash.no_update, '', cluster_err_alert, dash.no_update, dash.no_update, dash.no_update, \
                        text_score_cls_params, fig_obj_cluster_cls_params

            else:
                raise PreventUpdate

        @app.callback(
            Output('alert-to-reset-loss-cluster-cls', 'children'),
            Output('table-title-features-loss-cluster', 'children'),
            Output('show-feat-table-loss-cluster', 'children'),
            Output('table-title-probs-loss-cluster', 'children'),
            Output('show-prob-table-loss-cluster', 'children'),
            Input('fig-loss-cluster-cls', 'relayoutData'),
            Input('fig-loss-cluster-cls', 'restyleData'),
            State('select-misspred-dataset-cls', 'value'),
            State('select-num-cluster-cls', 'value'),
            State('select-log-method-cls', 'value'))
        def display_table_based_on_selected_range_cls(relayout_data, restyle_data, selected_dataset, selected_cluster, selected_method):
            '''
            Callbacks functionalities specific to reponse from fig-obj to data-table [ classification ]
            '''
            default_title = style_configs.DEFAULT_TITLE_STYLE
            title_table_features_cls = html.H6('Feature Values :', style=default_title, className='title__table-misspred-cls')
            title_table_probs_cls = html.H6('Probabilities Overview :', style=default_title, className='title__table-misspred-cls')

            if relayout_data is not None:
                specific_dataset = selected_dataset.replace('class ', '') if 'class' in selected_dataset else selected_dataset
                outputs_callback_fig_action = fig_plot_logloss_clusters_cls(self.data_loader,
                                                                        num_cluster=int(selected_cluster),
                                                                        log_func=style_configs.LOG_METHOD_DICT[selected_method],
                                                                        specific_dataset=specific_dataset)
                dfs_viz, df_features = outputs_callback_fig_action[0], outputs_callback_fig_action[6]

                try:
                    df_features = df_features.round(2)  # limit long decimals on feature values
                    dfs_viz[0] = dfs_viz[0].round(4)  # standardize prob values to 4 decimals
                    if self.is_bimodal:
                        dfs_viz[1] = dfs_viz[1].round(4)
                except TypeError:
                    df_features
                    dfs_viz

                df_features = insert_index_col(df_features)
                dfs_viz = [insert_index_col(df) for df in dfs_viz]

                if is_reset(relayout_data):
                    alert_obj_cls = None
                    title_table_features_cls = None
                    table_obj_features_cls = None
                    title_table_probs_cls = None
                    table_obj_probs_cls = None

                elif is_active_trace(relayout_data):
                    models = self.model_names
                    if restyle_data is not None:  # [{'visible': ['legendonly']}, [1]]
                        if detected_legend_filtration(restyle_data):
                            model_to_exclude_from_view = self.model_names[restyle_data[1][0]]
                            models = [model for model in self.model_names if model != model_to_exclude_from_view]

                    default_header = style_configs.default_header_style()
                    alert_obj_cls = style_configs.activate_alert()

                    # dfs_viz adjusted to the correct df according to the filtered model following click action on legend
                    dfs_viz_adjusted = get_adjusted_dfs_based_on_legend_filtration(dfs_viz, models)
                    df_final_features, df_final_probs = convert_cluster_relayout_data_to_df_cls(relayout_data,
                                                                                                dfs_viz_adjusted,
                                                                                                df_features,
                                                                                                models)

                    data_relayout_features_cls = df_final_features.to_dict('records')
                    data_relayout_prob_cls = df_final_probs.to_dict('recorfs')
                    table_obj_features_cls = table_with_relayout_datapoints(data_relayout_features_cls,
                                                                            df_final_features.columns,
                                                                            default_header,
                                                                            'csv')
                    table_obj_probs_cls = table_with_relayout_datapoints(data_relayout_prob_cls,
                                                                        df_final_probs.columns,
                                                                        default_header,
                                                                        'csv')
                return alert_obj_cls, title_table_features_cls, table_obj_features_cls, title_table_probs_cls, table_obj_probs_cls
            else:
                raise PreventUpdate



# Copyright 2021 AI Singapore. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict
import dash_table


def reponsive_table_to_filtered_datapoints(data: Dict, customized_cols: List[str], header: Dict, exp_format: str):
    '''
    Create table outlining dataframe content

    Arguments:
        data (:obj:`~dash_table.DataTable`):
            dictionary like format storing dataframe info under 'record' key
        customized_cols (:obj:`List[str]`):
            list of customized column names
        header (:obj:`Dict`):
            dictionary format storing the style info for table header
        exp_format (str):
            text info indicating the export format

    Returns:
        :obj:`~dash_table.DataTable`:
            table object outlining the dataframe content with specific styles
    '''
    table_obj = dash_table.DataTable(
                    data=data,
                    columns=[{'id': c, 'name': c} for c in customized_cols],
                    page_action='none',
                    fixed_rows={'headers': True},
                    fixed_columns={'headers': True, 'data': 1},
                    style_header=header,
                    style_data={'whiteSpace': 'normal', 'height': 'auto'},
                    style_cell={'textAlign': 'center',
                                'border': '1px solid rgb(229, 211, 197)',
                                'font-family': 'Arial',
                                'margin-bottom': '0',
                                'whiteSpace': 'normal',
                                'height': 'auto',
                                'minWidth': '200px',
                                'width': '200px',
                                'maxWidth': '200px'},
                    style_table={'width': 1400,
                                'height': 200,
                                'margin': 'auto'},
                    export_format=exp_format),
    return table_obj


def reponsive_table_to_filtered_datapoints_similaritiesCF(df, customized_cols, feature_cols, header, exp_format):
    '''
    Create table outlining dataframe content specific to Counter-Factuals component

    Arguments:
        df (:obj:`~pd.DataFrame`):
            dataframe containing calculated distance info
        customized_cols (:obj:`List[str]`):
            list of customized column names
        feature_cols (:obj:`List[str]`):
            list of feature column names
        header (:obj:`Dict`):
            dictionary format storing the style info for table header
        exp_format (str):
            text info indicating the export format

    Returns:
        :obj:`~dash_table.DataTable`:
            table object outlining the dataframe content with dynamic-conditional styles
    '''
    data = df.to_dict('records')
    table_obj = dash_table.DataTable(
                    data=data,
                    columns=[{'id': c, 'name': c} for c in customized_cols],
                    page_action='none',
                    fixed_rows={'headers': True},
                    fixed_columns={'headers': True, 'data': 1},
                    style_header=header,
                    style_data={'whiteSpace': 'normal', 'height': 'auto'},
                    style_cell={'textAlign': 'center',
                                'border': '1px solid rgb(229, 211, 197)',
                                'font-family': 'Arial',
                                'margin-bottom': '0',
                                'whiteSpace': 'normal',
                                'height': 'auto',
                                'minWidth': '200px',
                                'width': '200px',
                                'maxWidth': '200px'},
                    style_table={'width': 1400,
                                'height': 200,
                                'margin': 'auto'},
                    style_header_conditional=[
                        {
                            'if': {'column_id': col},
                            'backgroundColor': '#b69e8d',
                            'border': '1px solid white',
                        } for col in feature_cols
                    ],
                    style_data_conditional=[
                        {
                            'if': {'column_id': 'category'},
                            'border-top': '1px solid rgb(229, 211, 197)',
                            'fontWeight': 'bold'}
                    ] + [
                        {
                            'if': {'filter_query': '{category} = "User_defined_idx"'},
                            'backgroundColor': 'rgb(229, 240, 247)',
                        }
                    ],
                    export_format=exp_format),
    return table_obj
